""""""
import numpy
from scipy.linalg import block_diag, eig, inv, det
from math import pi

__all__ = ['amat', 'jmat', 'get_tunes_damp', 'get_mode_matrices', 'symplectify']

_i2 = numpy.array([[-1.j, -1.], [1., 1.j]])

# Prepare symplectic identity matrix
_j2 = numpy.array([[0., 1.], [-1., 0.]])
_jm = [_j2, block_diag(_j2, _j2), block_diag(_j2, _j2, _j2)]

_vxyz = [_i2, block_diag(_i2, _i2), block_diag(_i2, _i2, _i2)]
_submat = [slice(0, 2), slice(2, 4), slice(6, 3, -1)]


def jmat(ind):
    """
    Return the antisymetric block diagonal matrix [[0, 1][-1, 0]]

    INPUT
        ind     1, 2 or 3. Matrix dimension

    OUTPUT
        jm      block diagonal matrix, (2, 2) or (4, 4) or (6, 6)
    """
    return _jm[ind - 1]


def amat(tt):
    """
    find A matrix from one turn map matrix T such that:

               [Rotx  0    0  ]
    inv(A)*T*A=[ 0   Rotz  0  ]
               [ 0    0   Rots]

    Order it so that it is close to the order of x,y,z
    also ensure that positive modes are before negative so that
    one has proper symplecticity
    B. Nash July 18, 2013

    INPUT
    T       (m, m)  transfer matrix for 1 turn

    OUTPUT
    A       (m, m)  A-matrix

    """
    nv = tt.shape[0]
    dms = int(nv / 2)
    jmt = jmat(dms)
    select = numpy.arange(0, nv, 2)
    rbase = numpy.stack((select, select), axis=1).flatten()
    # noinspection PyTupleAssignmentBalance
    _, vv = eig(tt)
    # Compute the norms
    vp = numpy.dot(vv.conj().T, jmt)
    n = -0.5j * numpy.sum(vp.T * vv, axis=0)
    # Move positive before negatives
    order = rbase + (n < 0)
    vv = vv[:, order]
    n = n[order]
    # Normalize vectors
    vn = vv / numpy.sqrt(abs(n)).reshape((1, nv))
    # find the vectors that project most onto x,y,z, and reorder
    # nn will have structure
    #  n1x n1y n1z
    #  n2x n2y n2z
    #  n3x n3y n3z
    nn = 0.5 * abs(numpy.sqrt(-1.j * vn.conj().T.dot(jmt).dot(_vxyz[dms - 1])))
    rows = list(select)
    ord=[]
    for ixz in select:
        ind=numpy.argmax(nn[rows,ixz])
        ord.append(rows[ind])
        del rows[ind]
    v_ordered = vn[:,ord]
    aa = numpy.vstack((numpy.real(v_ordered), numpy.imag(v_ordered))).reshape(
        (nv, nv), order='F')
    return aa

def symplectify(M):
    """
    symplectify makes a matrix more symplectic
    follow Healy algorithm as described by McKay
    BNL-75461-2006-CP
    """
    J = jmat(3)

    V = numpy.dot(numpy.dot(J, numpy.identity(6) - M), numpy.linalg.inv(numpy.identity(6)+M))
    #V should be almost symmetric.  Replace with symmetrized version.

    W= ( V + V.T ) / 2
    #Now reconstruct M from W
    JW  = numpy.dot(J,W)
    MS=numpy.dot(numpy.identity(6) + JW, numpy.linalg.inv(numpy.identity(6)-JW))
    return MS


def get_mode_matrices(a):
    """Given a (m, m) A matrix , find the R-matrices of the m/2 normal modes"""
    dms = int(a.shape[0] / 2)
    return numpy.stack([numpy.dot(a[:, s], a.T[s, :]) for s in _submat[:dms]],
                       axis=0)


def get_tunes_damp(tt, rr=None):
    """
    mode_emit, damping_rates, tunes = get_tunes_damp(T, R)

    INPUT
        T                   (m, m) transfer matrix for 1 turn
        R                   (m, m) beam matrix (optional)

        m can be 2 (single plane), 4 (betatron motion) or 6 (full motion)

    OUTPUT
        record array with the following fields:
        tunes               (m/2,) tunes of the m/2 normal modes
        damping_rates       (m/2,) damping rates of the m/2 normal modes
        mode_matrices       (m/2, m, m) the R-matrices of the m/2 normal modes
        mode_emittances     Only if R is specified: (m/2,) emittance of each
                            of the m/2 normal modes
    """

    def decode(rot22):
        tune = (numpy.arctan2(rot22[0, 1] - rot22[1, 0],
                              rot22[0, 0] + rot22[1, 1]) / 2.0 / pi) % 1
        chi = -numpy.log(numpy.sqrt(det(rot22)))
        return chi, tune

    nv = tt.shape[0]
    dms = int(nv / 2)
    jmt = jmat(dms)
    aa = amat(tt)
    rmat = inv(aa).dot(tt.dot(aa))
    damping_rates, tunes = zip(*(decode(rmat[s, s]) for s in _submat[:dms]))
    if rr is None:
        return numpy.rec.fromarrays(
            (numpy.array(tunes), numpy.array(damping_rates),
             numpy.array(get_mode_matrices(aa))),
            dtype=[('tunes', numpy.float64, (dms,)),
                   ('damping_rates', numpy.float64, (dms,)),
                   ('mode_matrices', numpy.float64, (dms, nv, nv))]
        )
    else:
        rdiag = numpy.diag(aa.T.dot(jmt.dot(rr.dot(jmt.dot(aa)))))
        mode_emit = -0.5 * (rdiag[0:nv:2] + rdiag[1:nv:2])
        return numpy.rec.fromarrays(
            (numpy.array(tunes), numpy.array(damping_rates),
             numpy.array(get_mode_matrices(aa)), mode_emit),
            dtype=[('tunes', numpy.float64, (dms,)),
                   ('damping_rates', numpy.float64, (dms,)),
                   ('mode_matrices', numpy.float64, (dms, nv, nv)),
                   ('mode_emittances', numpy.float64, (dms,))]
        )



"""
Accelerator physics functions
"""

from math import sqrt, pi
from scipy.constants import physical_constants as cst

_e_radius = cst['classical electron radius'][0]
_hbar = cst['Planck constant over 2 pi times c in MeV fm'][0]

e_mass = 1.0e+06 * cst['electron mass energy equivalent in MeV'][0]  # eV
Cgamma = 4.0 * pi * _e_radius / 3.0 / pow(e_mass, 3)                 # m/eV^3
Cq = 55 / 32 / sqrt(3) * _hbar / e_mass * 1.0e-9                     # m

from .diff_constants import *
from .orbit import *
from .amat import *
from .matrix import *
from .linear import *
from .ring_parameters import *
# noinspection PyUnresolvedReferences
from .diffmatrix import find_mpole_raddiff_matrix
from .radiation import *
from .harmonic_analysis import *
from .nonlinear import *
from .fastring import *



"""
Closed orbit related functions
"""

import numpy
import scipy.constants as constants
from at.lattice import AtWarning, AtError, check_radiation
from at.lattice import Lattice, get_s_pos, elements, uint32_refpts
from at.tracking import lattice_pass
from at.physics import DConstant
import warnings

__all__ = ['find_orbit4', 'find_sync_orbit', 'find_orbit6']


@check_radiation(False)
def find_orbit4(ring, dp=0.0, refpts=None, guess=None, **kwargs):
    """findorbit4 finds the closed orbit in the 4-d transverse phase
    space by numerically solving for a fixed point of the one turn
    map M calculated with lattice_pass.

        (X, PX, Y, PY, dP, CT2 ) = M (X, PX, Y, PY, dP, CT1)

    under the CONSTANT MOMENTUM constraint dP and with NO constraint
    on the 6-th coordinate CT

    IMPORTANT!!! findorbit4 imposes a constraint on dP and relaxes
    the constraint on the revolution frequency. A physical storage
    ring does exactly the opposite: the momentum deviation of a
    particle on the closed orbit settles at the value
    such that the revolution is synchronous with the RF cavity

                HarmNumber*Frev = Frf

    To impose this artificial constraint in find_orbit4, PassMethod
    used for any element SHOULD NOT
    1.  change the longitudinal momentum dP (cavities , magnets with radiation)
    2.  have any time dependence (localized impedance, fast kickers etc)

    PARAMETERS
        ring            Sequence of AT elements
        dp              momentum deviation. Defaults to 0
        refpts          elements at which data is returned. It can be:
                        1) an integer in the range [-len(ring), len(ring)-1]
                           selecting the element according to python indexing
                           rules. As a special case, len(ring) is allowed and
                           refers to the end of the last element,
                        2) an ordered list of such integers without duplicates,
                        3) a numpy array of booleans of maximum length
                           len(ring)+1, where selected elements are True.
                        Defaults to None, if refpts is None an empty array is
                        returned for orbit.

    OUTPUT
        orbit0          ((6,) closed orbit vector at the entrance of the
                        1-st element (x,px,y,py)
        orbit           (6, Nrefs) closed orbit vector at each location
                        specified in refpts

    KEYWORDS
        keep_lattice    Assume no lattice change since the previous tracking.
                        Default: False
        guess           (6,) initial value for the closed orbit. It may help
                        convergence. Default: (0, 0, 0, 0, 0, 0)
        convergence     Convergence criterion. Default: 1.e-12
        max_iterations  Maximum number of iterations. Default: 20
        XYStep          Step size. Default: DConstant.XYStep

    See also find_sync_orbit, find_orbit6.
    """
    # We seek
    #  - f(x) = x
    #  - g(x) = f(x) - x = 0
    #  - g'(x) = f'(x) - 1
    # Use a Newton-Raphson-type algorithm:
    #  - r_n+1 = r_n - g(r_n) / g'(r_n)
    #  - r_n+1 = r_n - (f(r_n) - r_n) / (f'(r_n) - 1)
    #
    # (f(r_n) - r_n) / (f'(r_n) - 1) can be seen as x = b/a where we use least
    #     squares fitting to determine x when ax = b
    # f(r_n) - r_n is denoted b
    # f'(r_n) is the 4x4 jacobian, denoted j4

    keep_lattice = kwargs.pop('keep_lattice', False)
    convergence = kwargs.pop('convergence', DConstant.OrbConvergence)
    max_iterations = kwargs.pop('max_iterations', DConstant.OrbMaxIter)
    xy_step = kwargs.pop('XYStep', DConstant.XYStep)
    ref_in = numpy.zeros((6,), order='F') if guess is None else guess
    ref_in[4] = dp

    scaling = xy_step * numpy.array([1.0, 1.0, 1.0, 1.0])
    delta_matrix = numpy.zeros((6, 5), order='F')
    for i in range(4):
        delta_matrix[i, i] = scaling[i]
    id4 = numpy.asfortranarray(numpy.identity(4))
    change = 1
    itercount = 0
    while (change > convergence) and itercount < max_iterations:
        in_mat = ref_in.reshape((6, 1)) + delta_matrix
        _ = lattice_pass(ring, in_mat, refpts=[], keep_lattice=keep_lattice)
        # the reference particle after one turn
        ref_out = in_mat[:, 4]
        # 4x4 jacobian matrix from numerical differentiation:
        # f(x+d) - f(x) / d
        j4 = (in_mat[:4, :4] - in_mat[:4, 4:]) / scaling
        a = j4 - id4  # f'(r_n) - 1
        b = ref_out[:4] - ref_in[:4]
        b_over_a = numpy.linalg.solve(a, b)
        r_next = ref_in - numpy.append(b_over_a, numpy.zeros((2,)))
        # determine if we are close enough
        change = numpy.linalg.norm(r_next - ref_in)
        itercount += 1
        ref_in = r_next
        keep_lattice = True

    if itercount == max_iterations:
        warnings.warn(AtWarning('Maximum number of iterations reached.'
                                'Possible non-convergence'))

    uint32refs = uint32_refpts(refpts, len(ring))
    # bug in numpy < 1.13
    all_points = numpy.empty((0, 6), dtype=float) if len(
        uint32refs) == 0 else numpy.squeeze(
        lattice_pass(ring, ref_in.copy(order='K'), refpts=uint32refs,
                     keep_lattice=keep_lattice), axis=(1, 3)).T

    return ref_in, all_points


@check_radiation(False)
def find_sync_orbit(ring, dct=0.0, refpts=None, guess=None, **kwargs):
    """find_sync_orbit finds the closed orbit, synchronous with the RF cavity
    and momentum deviation dP (first 5 components of the phase space vector)
    % by numerically solving  for a fixed point
    % of the one turn map M calculated with lattice_pass

        (X, PX, Y, PY, dP, CT2 ) = M (X, PX, Y, PY, dP, CT1)

    under the constraint dCT = CT2 - CT1 = C/Frev - C/Frev0, where
    Frev0 = Frf0/HarmNumber is the design revolution frequency
    Frev  = (Frf0 + dFrf)/HarmNumber is the imposed revolution frequency

    IMPORTANT!!!  find_sync_orbit imposes a constraint (CT2 - CT1) and
    dP2 = dP1 but no constraint on the value of dP1, dP2
    The algorithm assumes time-independent fixed-momentum ring
    to reduce the dimensionality of the problem.

    To impose this artificial constraint in find_sync_orbit
    PassMethod used for any element SHOULD NOT
    1.  change the longitudinal momentum dP (cavities , magnets with radiation)
    2.  have any time dependence (localized impedance, fast kickers etc).

    PARAMETERS
        ring            Sequence of AT elements
        dct             Path lehgth deviation. Default: 0
        refpts          elements at which data is returned. It can be:
                        1) an integer in the range [-len(ring), len(ring)-1]
                           selecting the element according to python indexing
                           rules. As a special case, len(ring) is allowed and
                           refers to the end of the last element,
                        2) an ordered list of such integers without duplicates,
                        3) a numpy array of booleans of maximum length
                           len(ring)+1, where selected elements are True.
                        Defaults to None, if refpts is None an empty array is
                        returned for orbit.

    OUTPUT
        orbit0          ((6,) closed orbit vector at the entrance of the
                        1-st element (x,px,y,py)
        orbit           (6, Nrefs) closed orbit vector at each location
                        specified in refpts

    KEYWORDS
        keep_lattice    Assume no lattice change since the previous tracking.
                        Default: False
        guess           Initial value for the closed orbit. It may help
                        convergence. Default: (0, 0, 0, 0)
        convergence     Convergence criterion. Default: 1.e-12
        max_iterations  Maximum number of iterations. Default: 20
        XYStep          Step size. Default: DConstant.XYStep

    See also find_orbit4, find_orbit6.
    """
    keep_lattice = kwargs.pop('keep_lattice', False)
    convergence = kwargs.pop('convergence', DConstant.OrbConvergence)
    max_iterations = kwargs.pop('max_iterations', DConstant.OrbMaxIter)
    xy_step = kwargs.pop('XYStep', DConstant.XYStep)
    ref_in = numpy.zeros((6,), order='F') if guess is None else guess

    scaling = xy_step * numpy.array([1.0, 1.0, 1.0, 1.0, 1.0])
    delta_matrix = numpy.zeros((6, 6), order='F')
    for i in range(5):
        delta_matrix[i, i] = scaling[i]
    theta5 = numpy.zeros((5,), order='F')
    theta5[4] = dct
    id5 = numpy.zeros((5, 5), order='F')
    for i in range(4):
        id5[i, i] = 1.0
    idx = numpy.array([0, 1, 2, 3, 5])
    change = 1
    itercount = 0
    while (change > convergence) and itercount < max_iterations:
        in_mat = ref_in.reshape((6, 1)) + delta_matrix
        _ = lattice_pass(ring, in_mat, refpts=[], keep_lattice=keep_lattice)
        # the reference particle after one turn
        ref_out = in_mat[:, -1]
        # 5x5 jacobian matrix from numerical differentiation:
        # f(x+d) - f(x) / d
        j5 = (in_mat[idx, :5] - in_mat[idx, 5:]) / scaling
        a = j5 - id5  # f'(r_n) - 1
        b = ref_out[idx] - numpy.append(ref_in[:4], 0.0) - theta5
        b_over_a = numpy.linalg.solve(a, b)
        r_next = ref_in - numpy.append(b_over_a, 0.0)
        # determine if we are close enough
        change = numpy.linalg.norm(r_next - ref_in)
        itercount += 1
        ref_in = r_next
        keep_lattice = True

    if itercount == max_iterations:
        warnings.warn(AtWarning('Maximum number of iterations reached. '
                                'Possible non-convergence'))

    uint32refs = uint32_refpts(refpts, len(ring))
    # bug in numpy < 1.13
    all_points = numpy.empty((0, 6), dtype=float) if len(
        uint32refs) == 0 else numpy.squeeze(
        lattice_pass(ring, ref_in.copy(order='K'), refpts=uint32refs,
                     keep_lattice=keep_lattice), axis=(1, 3)).T

    return ref_in, all_points


def find_orbit6(ring, refpts=None, guess=None, **kwargs):
    """find_orbit6 finds the closed orbit in the full 6-D phase space
    by numerically solving  for a fixed point of the one turn
    map M calculated with lattice_pass

    (X, PX, Y, PY, DP, CT2 ) = M (X, PX, Y, PY, DP, CT1)

    with constraint  CT2 - CT1 = C*HarmNumber(1/Frf - 1/Frf0)

    IMPORTANT!!! find_orbit6 is a realistic simulation
    1.  The Frf frequency in the RF cavities (may be different from Frf0)
        imposes the synchronous condition
        CT2 - CT1 = C*HarmNumber(1/Frf - 1/Frf0)
    2.  The algorithm numerically calculates
        6-by-6 Jacobian matrix J6. In order for (J-E) matrix
        to be non-singular it is NECESSARY to use a realistic
        PassMethod for cavities with non-zero momentum kick
        (such as ThinCavityPass).
    3.  find_orbit6 can find orbits with radiation.
        In order for the solution to exist the cavity must supply
        adequate energy compensation.
        In the simplest case of a single cavity, it must have
        'Voltage' field set so that Voltage > Erad - energy loss per turn
    4.  There is a family of solutions that correspond to different RF buckets
        They differ in the 6-th coordinate by C*Nb/Frf. Nb = 1 .. HarmNum-1
    5.  The value of the 6-th coordinate found at the cavity gives
        the equilibrium RF phase. If there is no radiation the phase is 0;

    PARAMETERS
        ring            Sequence of AT elements
        refpts          elements at which data is returned. It can be:
                        1) an integer in the range [-len(ring), len(ring)-1]
                           selecting the element according to python indexing
                           rules. As a special case, len(ring) is allowed and
                           refers to the end of the last element,
                        2) an ordered list of such integers without duplicates,
                        3) a numpy array of booleans of maximum length
                           len(ring)+1, where selected elements are True.
                        Defaults to None, if refpts is None an empty array is
                        returned for orbit.

    OUTPUT
        orbit0          ((6,) closed orbit vector at the entrance of the
                        1-st element (x,px,y,py)
        orbit           (6, Nrefs) closed orbit vector at each location
                        specified in refpts

    KEYWORDS
        keep_lattice    Assume no lattice change since the previous tracking.
                        Default: False
        guess           Initial value for the closed orbit. It may help
                        convergence. Default: (0, 0, 0, 0)
        convergence     Convergence criterion. Default: 1.e-12
        max_iterations  Maximum number of iterations. Default: 20
        XYStep          Step size. Default: DConstant.XYStep
        DPStep          Step size. Default: DConstant.XYStep

    See also find_orbit4, find_sync_orbit.
    """
    keep_lattice = kwargs.pop('keep_lattice', False)
    convergence = kwargs.pop('convergence', DConstant.OrbConvergence)
    max_iterations = kwargs.pop('max_iterations', DConstant.OrbMaxIter)
    xy_step = kwargs.pop('XYStep', DConstant.XYStep)
    dp_step = kwargs.pop('DPStep', DConstant.DPStep)
    ref_in = numpy.zeros((6,), order='F') if guess is None else guess

    # Get evolution period
    l0 = get_s_pos(ring, len(ring))
    cavities = [elem for elem in ring if isinstance(elem, elements.RFCavity)]
    if len(cavities) == 0:
        raise AtError('No cavity found in the lattice.')

    f_rf = cavities[0].Frequency
    harm_number = cavities[0].HarmNumber

    theta = numpy.zeros((6,))
    theta[5] = constants.speed_of_light * harm_number / f_rf - l0

    scaling = xy_step * numpy.array([1.0, 1.0, 1.0, 1.0, 0.0, 0.0]) + \
              dp_step * numpy.array([0.0, 0.0, 0.0, 0.0, 1.0, 1.0])
    delta_matrix = numpy.asfortranarray(
        numpy.concatenate((numpy.diag(scaling), numpy.zeros((6, 1))), axis=1))

    id6 = numpy.asfortranarray(numpy.identity(6))
    change = 1
    itercount = 0
    while (change > convergence) and itercount < max_iterations:
        in_mat = ref_in.reshape((6, 1)) + delta_matrix
        _ = lattice_pass(ring, in_mat, refpts=[], keep_lattice=keep_lattice)
        # the reference particle after one turn
        ref_out = in_mat[:, 6]
        # 6x6 jacobian matrix from numerical differentiation:
        # f(x+d) - f(x) / d
        j6 = (in_mat[:, :6] - in_mat[:, 6:]) / scaling
        a = j6 - id6  # f'(r_n) - 1
        b = ref_out[:] - ref_in[:] - theta
        # b_over_a, _, _, _ = numpy.linalg.lstsq(a, b, rcond=-1)
        b_over_a = numpy.linalg.solve(a, b)
        r_next = ref_in - b_over_a
        # determine if we are close enough
        change = numpy.linalg.norm(r_next - ref_in)
        itercount += 1
        ref_in = r_next
        keep_lattice = True

    if itercount == max_iterations:
        warnings.warn(AtWarning('Maximum number of iterations reached. '
                                'Possible non-convergence'))

    uint32refs = uint32_refpts(refpts, len(ring))
    # bug in numpy < 1.13
    all_points = numpy.empty((0, 6), dtype=float) if len(
        uint32refs) == 0 else numpy.squeeze(
        lattice_pass(ring, ref_in.copy(order='K'), refpts=uint32refs,
                     keep_lattice=keep_lattice), axis=(1, 3)).T

    return ref_in, all_points


Lattice.find_orbit4 = find_orbit4
Lattice.find_sync_orbit = find_sync_orbit
Lattice.find_orbit6 = find_orbit6



import numpy
from at.tracking import atpass, elempass
from at.lattice import uint32_refpts


__all__ = ['lattice_pass', 'element_pass']

DIMENSION_ERROR = 'Input to lattice_pass() must be a 6xN array.'


def lattice_pass(lattice, r_in, nturns=1, refpts=None, keep_lattice=False):
    """lattice_pass tracks particles through each element of a lattice
    calling the element-specific tracking function specified in the
    lattice[i].PassMethod field.

    Note:

     * lattice_pass(lattice, r_in, refpts=len(line)) is the same as
       lattice_pass(lattice, r_in) since the reference point len(line) is the
       exit of the last element
     * lattice_pass(lattice, r_in, refpts=0) is a copy of r_in since the
       reference point 0 is the entrance of the first element

    PARAMETERS
        lattice:    iterable of AT elements
        r_in:       (6, N) array: input coordinates of N particles.
                    r_in is modified in-place and reports the coordinates at
                    the end of the tracking. For the the best efficiency, r_in
                    should be given as F_CONTIGUOUS numpy array.
        nturns:     number of passes through the lattice line
        refpts      elements at which data is returned. It can be:
                    1) an integer in the range [-len(ring), len(ring)-1]
                       selecting the element according to python indexing
                       rules. As a special case, len(ring) is allowed and
                       refers to the end of the last element,
                    2) an ordered list of such integers without duplicates,
                    3) a numpy array of booleans of maximum length
                       len(ring)+1, where selected elements are True.
                    Defaults to None, meaning no refpts, equivelent to
                    passing an empty array for calculation purposes.
        keep_lattice: use elements persisted from a previous call to at.atpass.
                    If True, assume that the lattice has not changed since
                    that previous call.

    OUTPUT
        (6, N, R, T) array containing output coordinates of N particles
        at R reference points for T turns.
    """
    assert r_in.shape[0] == 6 and r_in.ndim in (1, 2), DIMENSION_ERROR
    if not isinstance(lattice, list):
        lattice = list(lattice)
    nelems = len(lattice)
    if refpts is None:
        refpts = nelems
    refs = uint32_refpts(refpts, nelems)
    # atpass returns 6xAxBxC array where n = x*y*z;
    # * A is number of particles;
    # * B is number of refpts
    # * C is the number of turns
    if r_in.flags.f_contiguous:
        return atpass(lattice, r_in, nturns, refs, int(keep_lattice))
    else:
        r_fin = numpy.asfortranarray(r_in)
        r_out = atpass(lattice, r_fin, nturns, refs, int(keep_lattice))
        r_in[:] = r_fin[:]
        return r_out


def element_pass(element, r_in):
    r_in = numpy.asfortranarray(r_in)
    elempass(element, r_in)



"""AT generic plotting function"""
from itertools import chain, repeat
# noinspection PyPackageRequirements
import matplotlib.pyplot as plt
from at.plot import plot_synopt

SLICES = 400

__all__ = ['baseplot']


def baseplot(ring, plot_function, *args, **kwargs):
    """
    baseplot divides the region of interest of ring into small elements,
    calls the specified function to get the plot data and calls matplotlib
    functions to generate the plot.
    By default it creates a new figure for the plot, but if provided with
    axes objects it can be used as part of a GUI

    PARAMETERS
        ring            Lattice object
        plot_function   specific data generating function to be called

        All other positional parameters are sent to the plotting function

        plot_function is called as:

        title, left, right = plot_function(ring, refpts, *args, **kwargs)

        and should return 2 or 3 output:

        title   plot title or None
        left    tuple returning the data for the main (left) axis
           left[0]   y-axis label
           left[1]   xdata: (N,) array (s coordinate)
           left[2]   ydata: iterable of (N,) or (N,M) arrays. Lines from a
                     (N, M) array share the same style and label
           left[3]   labels: (optional) iterable of strings as long as ydata
        right   tuple returning the data for the secondary (right) axis
                (optional)

    KEYWORDS
        s_range         lattice range of interest, default: unchanged,
                        initially set to the full cell.
        axes=None       axes for plotting as (primary_axes, secondary_axes)
                        Default: create new axes
        slices=400      Number of slices
        legend=True     Show a legend on the plot
        block=False     if True, block until the figure is closed
        dipole={}       Dictionary of properties overloading the default
                        properties of dipole representation.
                        See 'plot_synopt' for details
        quadrupole={}   Same definition as for dipole
        sextupole={}    Same definition as for dipole
        multipole={}    Same definition as for dipole
        monitor={}      Same definition as for dipole

        All other keywords are sent to the plotting function

    RETURN
        left_axes       Main (left) axes
        right_axes      Secondary (right) axes or None
        synopt_axes     Synoptic axes
    """

    def plot1(ax, yaxis_label, x, y, labels=()):
        lines = []
        for y1, prop, label in zip(y, props, chain(labels, repeat(None))):
            ll = ax.plot(x, y1, **prop)
            if label is not None:
                ll[0].set_label(label)
            lines += ll
        ax.set_ylabel(yaxis_label)
        return lines

    def labeled(line):
        return not line.properties()['label'].startswith('_')

    # extract baseplot arguments
    slices = kwargs.pop('slices', SLICES)
    axes = kwargs.pop('axes', None)
    legend = kwargs.pop('legend', True)
    block = kwargs.pop('block', False)
    if 's_range' in kwargs:
        ring.s_range = kwargs.pop('s_range')

    # extract synopt arguments
    synkeys = ['dipole', 'quadrupole', 'sextupole', 'multipole', 'monitor']
    kwkeys = list(kwargs.keys())
    synargs = dict((k, kwargs.pop(k)) for k in kwkeys if k in synkeys)

    # get color cycle
    cycle_props = plt.rcParams['axes.prop_cycle']

    # slice the ring
    rg = ring.slice(slices=slices)

    # get the data for the plot
    pout = plot_function(rg, rg.i_range, *args, **kwargs)
    title = pout[0]
    plots = pout[1:]

    # prepare the axes
    if axes is None:
        # Create new axes
        nplots = len(plots)
        fig = plt.figure()
        axleft = fig.add_subplot(111, xlim=rg.s_range, xlabel='s [m]',
                                 facecolor=[1.0, 1.0, 1.0, 0.0],
                                 title=title)
        axright = axleft.twinx() if (nplots >= 2) else None
        axleft.set_title(ring.name, fontdict={'fontsize': 'medium'},
                         loc='left')
        axsyn = plot_synopt(ring, axes=axleft, **synargs)
    else:
        # Use existing axes
        axleft, axright = axes
        axsyn = None
        nplots = 1 if axright is None else len(plots)

    props = iter(cycle_props())

    # left plot
    lines1 = plot1(axleft, *plots[0])
    # right plot
    lines2 = [] if (nplots < 2) else plot1(axright, *plots[1])
    if legend:
        if nplots < 2:
            axleft.legend(handles=[l for l in lines1 if labeled(l)])
        elif axleft.get_shared_x_axes().joined(axleft, axright):
            axleft.legend(
                handles=[l for l in lines1 + lines2 if labeled(l)])
        else:
            axleft.legend(handles=[l for l in lines1 if labeled(l)])
            axright.legend(handles=[l for l in lines2 if labeled(l)])
    plt.show(block=block)
    return axleft, axright, axsyn



"""
Simple parallelisation of atpass() using multiprocessing.
"""
import multiprocessing
from at.tracking import atpass
from at.lattice import uint32_refpts
import numpy


__all__ = ['patpass']


def _atpass_one(args):
    ring, rin, turns, refpts = args
    return atpass(ring, rin, turns, refpts)


def _patpass(ring, r_in, nturns, refpts, pool_size):
    pool = multiprocessing.Pool(pool_size)
    args = [(ring, r_in[:, i], nturns, refpts) for i in range(r_in.shape[1])]
    results = pool.map(_atpass_one, args)
    return numpy.concatenate(results, axis=1)


def patpass(ring, r_in, nturns, refpts=None, reuse=True, pool_size=None):
    """
    Simple parallel implementation of atpass().  If more than one particle
    is supplied, use multiprocessing to run each particle in a separate
    process.

    INPUT:
        ring            lattice description
        r_in:           6xN array: input coordinates of N particles
        nturns:         number of passes through the lattice line
        refpts          elements at which data is returned. It can be:
                        1) an integer in the range [-len(ring), len(ring)-1]
                           selecting the element according to python indexing
                           rules. As a special case, len(ring) is allowed and
                           refers to the end of the last element,
                        2) an ordered list of such integers without duplicates,
                        3) a numpy array of booleans of maximum length
                           len(ring)+1, where selected elements are True.
                        Defaults to None, meaning no refpts, equivelent to
                        passing an empty array for calculation purposes.
    """
    if not reuse:
        raise ValueError('patpass does not support altering lattices')
    if refpts is None:
        refpts = len(ring)
    refs = uint32_refpts(refpts, len(ring))
    if pool_size is None:
        pool_size = multiprocessing.cpu_count()
    return _patpass(ring, r_in, nturns, refs, pool_size)



"""
transfer matrix related functions

A collection of functions to compute 4x4 and 6x6 transfer matrices
"""

import numpy
from at.lattice import Lattice, uint32_refpts, get_refpts
from at.lattice.elements import Bend, M66
from at.tracking import lattice_pass, element_pass
from at.physics import find_orbit4, find_orbit6, jmat, symplectify, DConstant

__all__ = ['find_m44', 'find_m66', 'find_elem_m66', 'gen_m66_elem']

_jmt = jmat(2)


def find_m44(ring, dp=0.0, refpts=None, orbit=None, keep_lattice=False,
             **kwargs):
    """find_m44 numerically finds the 4x4 transfer matrix of an accelerator
    lattice for a particle with relative momentum deviation DP

    IMPORTANT!!! find_m44 assumes constant momentum deviation.
    PassMethod used for any element in the lattice SHOULD NOT
    1.  change the longitudinal momentum dP
        (cavities , magnets with radiation, ...)
    2.  have any time dependence (localized impedance, fast kickers, ...)

    m44, t = find_m44(lattice, dp=0.0, refpts)
        return 4x4 transfer matrices between the entrance of the first element
        and each element indexed by refpts.
            m44:    full one-turn matrix at the entrance of the first element
            t:      4x4 transfer matrices between the entrance of the first
                    element and each element indexed by refpts:
                    (Nrefs, 4, 4) array

    Unless an input orbit is introduced, find_m44 assumes that the lattice is
    a ring and first finds the closed orbit.

    PARAMETERS
        ring            lattice description
        dp              momentum deviation. Defaults to 0
        refpts          elements at which data is returned. It can be:
                        1) an integer in the range [-len(ring), len(ring)-1]
                           selecting the element according to python indexing
                           rules. As a special case, len(ring) is allowed and
                           refers to the end of the last element,
                        2) an ordered list of such integers without duplicates,
                        3) a numpy array of booleans of maximum length
                           len(ring)+1, where selected elements are True.
                        Defaults to None, if refpts is None an empty array is
                        returned for mstack.

    KEYWORDS
        keep_lattice=False  When True, assume no lattice change since the
                            previous tracking.
        full=False          When True, matrices are full 1-turn matrices at
                            the entrance of each
                            element indexed by refpts.
        orbit=None          Avoids looking for the closed orbit if is already
                            known (6,) array
        XYStep=1.e-8        transverse step for numerical computation

    See also find_m66, find_orbit4
    """

    def mrotate(m):
        m = numpy.squeeze(m)
        return m.dot(m44.dot(_jmt.T.dot(m.T.dot(_jmt))))

    xy_step = kwargs.pop('XYStep', DConstant.XYStep)
    full = kwargs.pop('full', False)
    if orbit is None:
        orbit, _ = find_orbit4(ring, dp, keep_lattice=keep_lattice,
                               XYStep=xy_step)
        keep_lattice = True
    # Construct matrix of plus and minus deltas
    # scaling = 2*xy_step*numpy.array([1.0, 0.1, 1.0, 0.1])
    scaling = xy_step * numpy.array([1.0, 1.0, 1.0, 1.0])
    dg = numpy.asfortranarray(
        numpy.concatenate((0.5 * numpy.diag(scaling), numpy.zeros((2, 4)))))
    dmat = numpy.concatenate((dg, -dg), axis=1)
    # Add the deltas to multiple copies of the closed orbit
    in_mat = orbit.reshape(6, 1) + dmat

    refs = uint32_refpts(refpts, len(ring))
    out_mat = numpy.rollaxis(
        numpy.squeeze(lattice_pass(ring, in_mat, refpts=refs,
                                   keep_lattice=keep_lattice), axis=3), -1
    )
    # out_mat: 8 particles at n refpts for one turn
    # (x + d) - (x - d) / d
    m44 = (in_mat[:4, :4] - in_mat[:4, 4:]) / scaling

    if len(refs) > 0:
        mstack = (out_mat[:, :4, :4] - out_mat[:, :4, 4:]) / scaling
        if full:
            mstack = numpy.stack([mrotate(mat) for mat in mstack], axis=0)
    else:
        mstack = numpy.empty((0, 4, 4), dtype=float)

    return m44, mstack


def find_m66(ring, refpts=None, orbit=None, keep_lattice=False, **kwargs):
    """find_m66 numerically finds the 6x6 transfer matrix of an accelerator
    lattice by differentiation of lattice_pass near the closed orbit.
    find_m66 uses find_orbit6 to search for the closed orbit in 6-D
    In order for this to work the ring MUST have a CAVITY element

    m66, t = find_m66(lattice, refpts)
        m66:    full one-turn 6-by-6 matrix at the entrance of the
                first element.
        t:      6x6 transfer matrices between the entrance of the first
                element and each element indexed by refpts (nrefs, 6, 6) array.

    PARAMETERS
        ring            lattice description
        dp              momentum deviation. Defaults to 0
        refpts          elements at which data is returned. It can be:
                        1) an integer in the range [-len(ring), len(ring)-1]
                           selecting the element according to python indexing
                           rules. As a special case, len(ring) is allowed and
                           refers to the end of the last element,
                        2) an ordered list of such integers without duplicates,
                        3) a numpy array of booleans of maximum length
                           len(ring)+1, where selected elements are True.
                        Defaults to None, if refpts is None an empty array is
                        returned for mstack.

    KEYWORDS
        keep_lattice=False  When True, assume no lattice change since the
                            previous tracking.
        orbit=None          Avoids looking for the closed orbit if is already
                            known (6,) array
        XYStep=1.e-8        transverse step for numerical computation

    See also find_m44, find_orbit6
    """
    xy_step = kwargs.pop('XYStep', DConstant.XYStep)
    dp_step = kwargs.pop('DPStep', DConstant.DPStep)
    if orbit is None:
        orbit, _ = find_orbit6(ring, keep_lattice=keep_lattice,
                               XYStep=xy_step, DPStep=dp_step)
        keep_lattice = True

    # Construct matrix of plus and minus deltas
    # scaling = 2*xy_step*numpy.array([1.0, 0.1, 1.0, 0.1, 1.0, 1.0])
    scaling = xy_step * numpy.array([1.0, 1.0, 1.0, 1.0, 0.0, 0.0]) + \
              dp_step * numpy.array([0.0, 0.0, 0.0, 0.0, 1.0, 1.0])
    dg = numpy.asfortranarray(0.5 * numpy.diag(scaling))
    dmat = numpy.concatenate((dg, -dg), axis=1)

    in_mat = orbit.reshape(6, 1) + dmat

    refs = uint32_refpts(refpts, len(ring))
    out_mat = numpy.rollaxis(
        numpy.squeeze(lattice_pass(ring, in_mat, refpts=refs,
                                   keep_lattice=keep_lattice), axis=3), -1
    )
    # out_mat: 12 particles at n refpts for one turn
    # (x + d) - (x - d) / d
    m66 = (in_mat[:, :6] - in_mat[:, 6:]) / scaling

    if len(refs) > 0:
        mstack = (out_mat[:, :, :6] - out_mat[:, :, 6:]) / scaling
    else:
        mstack = numpy.empty((0, 6, 6), dtype=float)

    return m66, mstack


def find_elem_m66(elem, orbit=None, **kwargs):
    """
    Numerically find the 6x6 transfer matrix of a single element

    INPUT
        elem                AT element

    KEYWORDS
        orbit=None          closed orbit at the entrance of the element,
                            default: 0.0
        XYStep=1.e-8        transverse step for numerical computation

    OUTPUT
        m66                 (6, 6) transfer matrix
    """
    xy_step = kwargs.pop('XYStep', DConstant.XYStep)
    if orbit is None:
        orbit = numpy.zeros((6,))

    # Construct matrix of plus and minus deltas
    # scaling = 2*xy_step*numpy.array([1.0, 0.1, 1.0, 0.1, 1.0, 1.0])
    scaling = xy_step * numpy.array([1.0, 1.0, 1.0, 1.0, 1.0, 1.0])
    dg = numpy.asfortranarray(0.5 * numpy.diag(scaling))
    dmat = numpy.concatenate((dg, -dg), axis=1)

    in_mat = orbit.reshape(6, 1) + dmat
    element_pass(elem, in_mat)
    m66 = (in_mat[:, :6] - in_mat[:, 6:]) / scaling
    return m66


def gen_m66_elem(ring, o4b, o4e):
    """
    converts a ring to a linear 6x6 matrix tracking elemtn
    """

    dip_inds = get_refpts(ring, Bend)
    theta = numpy.array([ring[ind].BendingAngle for ind in dip_inds])
    lendp = numpy.array([ring[ind].Length for ind in dip_inds])
    s_pos = ring.get_s_pos()
    s = numpy.diff(numpy.array([s_pos[0], s_pos[-1]]))[0]
    i2 = numpy.sum(numpy.abs(theta * theta / lendp))
    m66_mat, _ = find_m66(ring, [], o4b)
    if ring.radiation is False:
        m66_mat = symplectify(m66_mat)  # remove for damping
    lin_elem = M66('Linear', m66_mat, T1=-o4b, T2=o4e, Length=s, I2=i2)
    return lin_elem


Lattice.find_m44 = find_m44
Lattice.find_m66 = find_m66



"""
Module to define common elements used in AT.

Each element has a default PassMethod attribute for which it should have the
appropriate attributes.  If a different PassMethod is set, it is the caller's
responsibility to ensure that the appropriate attributes are present.
"""
import re
import numpy
import copy
from inspect import getmembers, isdatadescriptor


def _array(value, shape=(-1,), dtype=numpy.float64):
    # Ensure proper ordering(F) and alignment(A) for "C" access in integrators
    return numpy.require(value, dtype=dtype, requirements=['F', 'A']).reshape(
        shape, order='F')


def _array66(value):
    return _array(value, shape=(6, 6))


def _nop(value):
    return value


class Element(object):
    """Base of pyat elements"""

    REQUIRED_ATTRIBUTES = ['FamName']
    _conversions = dict(FamName=str, PassMethod=str, Length=float,
                        R1=_array66, R2=_array66,
                        T1=lambda v: _array(v, (6,)),
                        T2=lambda v: _array(v, (6,)),
                        RApertures=lambda v: _array(v, (4,)),
                        EApertures=lambda v: _array(v, (2,)),
                        KickAngle=lambda v: _array(v, (2,)),
                        PolynomB=_array, PolynomA=_array,
                        BendingAngle=float,
                        MaxOrder=int, NumIntSteps=int,
                        Energy=float,
                        )

    _entrance_fields = ['T1', 'R1']
    _exit_fields = ['T2', 'R2']

    def __init__(self, family_name, **kwargs):
        self.FamName = family_name
        self.Length = kwargs.pop('Length', 0.0)
        self.PassMethod = kwargs.pop('PassMethod', 'IdentityPass')
        self.update(kwargs)

    def __setattr__(self, key, value):
        try:
            super(Element, self).__setattr__(
                key, self._conversions.get(key, _nop)(value))
        except Exception as exc:
            exc.args = ('In element {0}, parameter {1}: {2}'.format(
                self.FamName, key, exc),)
            raise

    def __str__(self):
        first3 = ['FamName', 'Length', 'PassMethod']
        attrs = dict(self.items())
        keywords = ['\t{0} : {1!s}'.format(k, attrs.pop(k)) for k in first3]
        keywords += ['\t{0} : {1!s}'.format(k, v) for k, v in attrs.items()]
        return '\n'.join((self.__class__.__name__ + ':', '\n'.join(keywords)))

    def __repr__(self):
        attrs = dict(self.items())
        arguments = [attrs.pop(k, getattr(self, k)) for k in
                     self.REQUIRED_ATTRIBUTES]
        defelem = self.__class__(*arguments)
        keywords = ['{0!r}'.format(arg) for arg in arguments]
        keywords += ['{0}={1!r}'.format(k, v) for k, v in sorted(attrs.items())
                     if not numpy.array_equal(v, getattr(defelem, k, None))]
        args = re.sub('\n\s*', ' ', ', '.join(keywords))
        return '{0}({1})'.format(self.__class__.__name__, args)

    def equals(self, other):
        """Whether an element is equivalent to another.

        This implementation was found to be too slow for the generic
        __eq__ method when comparing lattices.

        """
        return repr(self) == repr(other)

    def divide(self, frac):
        """split the element in len(frac) pieces whose length
        is frac[i]*self.Length

        arguments:
            frac            length of each slice expressed as a fraction of the
                            initial length. sum(frac) may differ from 1.

        Return a list of elements equivalent to the original.

        Example:

        >>> Drift('dr', 0.5).divide([0.2, 0.6, 0.2])
        [Drift('dr', 0.1), Drift('dr', 0.3), Drift('dr', 0.1)]
        """
        # Bx default, the element is indivisible
        return [self]

    def update(self, *args, **kwargs):
        """Update the element attributes with the given arguments

        update(**kwargs)
        update(mapping, **kwargs)
        update(iterable, **kwargs)
        """
        attrs = dict(*args, **kwargs)
        for (key, value) in attrs.items():
            setattr(self, key, value)

    def copy(self):
        """Return a shallow copy of the element"""
        return copy.copy(self)

    def deepcopy(self):
        """Return a deep copy of the element"""
        return copy.deepcopy(self)

    def items(self):
        """Iterates through the data members including slots and properties"""
        # Get attributes
        for k, v in vars(self).items():
            yield k, v
        # Get slots and properties
        for k, v in getmembers(self.__class__, isdatadescriptor):
            if not k.startswith('_'):
                yield k, getattr(self, k)


class LongElement(Element):
    """pyAT long element"""
    REQUIRED_ATTRIBUTES = Element.REQUIRED_ATTRIBUTES + ['Length']

    def __init__(self, family_name, length, *args, **kwargs):
        kwargs.setdefault('Length', length)
        # Ancestor may be either Element of ThinMultipole
        # noinspection PyArgumentList
        super(LongElement, self).__init__(family_name, *args, **kwargs)

    def _part(self, fr, sumfr):
        pp = self.copy()
        pp.Length = fr * self.Length
        if hasattr(self, 'KickAngle'):
            pp.KickAngle = fr / sumfr * self.KickAngle
        return pp

    def divide(self, frac):
        """split the element in len(frac) pieces whose length
        is frac[i]*self.Length

        arguments:
            frac            length of each slice expressed as a fraction of the
                            initial length. sum(frac) may differ from 1.

        Return a list of elements equivalent to the original.

        Example:

        >>> Drift('dr', 0.5).divide([0.2, 0.6, 0.2])
        [Drift('dr', 0.1), Drift('dr', 0.3), Drift('dr', 0.1)]
        """

        def popattr(element, attr):
            val = getattr(element, attr)
            delattr(element, attr)
            return attr, val

        frac = numpy.asarray(frac, dtype=float)
        el = self.copy()
        # Remove entrance and exit attributes
        fin = dict(popattr(el, key) for key in vars(self) if
                   key in self._entrance_fields)
        fout = dict(popattr(el, key) for key in vars(self) if
                    key in self._exit_fields)
        # Split element
        element_list = [el._part(f, numpy.sum(frac)) for f in frac]
        # Restore entrance and exit attributes
        for key, value in fin.items():
            setattr(element_list[0], key, value)
        for key, value in fout.items():
            setattr(element_list[-1], key, value)
        return element_list


class Marker(Element):
    """pyAT marker element"""


class Monitor(Element):
    """pyAT monitor element"""


class Aperture(Element):
    """pyAT aperture element"""
    REQUIRED_ATTRIBUTES = Element.REQUIRED_ATTRIBUTES + ['Limits']
    _conversions = dict(Element._conversions, Limits=lambda v: _array(v, (4,)))

    def __init__(self, family_name, limits, **kwargs):
        kwargs.setdefault('PassMethod', 'AperturePass')
        super(Aperture, self).__init__(family_name, Limits=limits, **kwargs)


class Drift(LongElement):
    """pyAT drift space element"""

    def __init__(self, family_name, length, **kwargs):
        """Drift(FamName, Length, **keywords)
        """
        kwargs.setdefault('PassMethod', 'DriftPass')
        super(Drift, self).__init__(family_name, length, **kwargs)

    def insert(self, insert_list):
        """insert elements inside a drift

        arguments:
            insert_list: iterable, each item of insert_list is itself an
                         iterable with 2 objects:
                             1. the location where the center of the element
                                will be inserted, given as a fraction of the
                                Drift length.
                             2. an element to be inserted at that location. If
                                None, the drift will be divided but no element
                                will be inserted.

        Return a list of elements.

        Drifts with negative lengths may be generated if necessary.

        Examples:

        >>> Drift('dr', 2.0).insert(((0.25, None), (0.75, None)))
        [Drift('dr', 0.5), Drift('dr', 1.0), Drift('dr', 0.5)]

        >>> Drift('dr', 2.0).insert(((0.0, Marker('m1')), (0.5, Marker('m2'))))
        [Marker('m1'), Drift('dr', 1.0), Marker('m2'), Drift('dr', 1.0)]

        >>> Drift('dr', 2.0).insert(((0.5, Quadrupole('qp', 0.4, 0.0)),))
        [Drift('dr', 0.8), Quadrupole('qp', 0.4), Drift('dr', 0.8)]
        """
        frac, elements = zip(*insert_list)
        lg = [0.0 if el is None else el.Length for el in elements]
        fr = numpy.asarray(frac, dtype=float)
        lg = 0.5 * numpy.asarray(lg, dtype=float) / self.Length
        drfrac = numpy.hstack((fr - lg, 1.0)) - numpy.hstack((0.0, fr + lg))
        long_elems = (drfrac != 0.0)
        drifts = numpy.ndarray((len(drfrac),), dtype='O')
        drifts[long_elems] = self.divide(drfrac[long_elems])
        line = [None] * (len(drifts) + len(elements))
        line[::2] = drifts
        line[1::2] = elements
        return [el for el in line if el is not None]


class ThinMultipole(Element):
    """pyAT thin multipole element"""
    REQUIRED_ATTRIBUTES = Element.REQUIRED_ATTRIBUTES + ['PolynomA',
                                                         'PolynomB']

    def __init__(self, family_name, poly_a, poly_b, **kwargs):
        """ThinMultipole(FamName, PolynomA, PolynomB, **keywords)

        Available keywords:
        MaxOrder        Number of desired multipoles. Default: highest index of
                        non-zero polynomial coefficients
        """
        def getpol(poly):
            nonzero = numpy.flatnonzero(poly != 0.0)
            return poly, len(poly), nonzero[-1] if len(nonzero) > 0 else -1

        def lengthen(poly, dl):
            if dl > 0:
                return numpy.concatenate((poly, numpy.zeros(dl)))
            else:
                return poly

        # Remove MaxOrder, PolynomA and PolynomB
        poly_a, len_a, ord_a = getpol(_array(kwargs.pop('PolynomA', poly_a)))
        poly_b, len_b, ord_b = getpol(_array(kwargs.pop('PolynomB', poly_b)))
        deforder = max(getattr(self, 'DefaultOrder', 0), ord_a, ord_b)
        maxorder = kwargs.pop('MaxOrder', deforder)
        len_ab = max(maxorder + 1, len_a, len_b)
        kwargs.setdefault('PassMethod', 'ThinMPolePass')
        super(ThinMultipole, self).__init__(family_name, **kwargs)
        # Set MaxOrder while PolynomA and PolynomB are not set yet
        super(ThinMultipole, self).__setattr__('MaxOrder', maxorder)
        # Adjust polynom lengths and set them
        self.PolynomA = lengthen(poly_a, len_ab - len_a)
        self.PolynomB = lengthen(poly_b, len_ab - len_b)

    def __setattr__(self, key, value):
        """Check the compatibility of MaxOrder, PolynomA and PolynomB"""
        polys = ('PolynomA', 'PolynomB')
        if key in polys:
            value = _array(value)
            lmin = getattr(self, 'MaxOrder')
            if not len(value) > lmin:
                raise ValueError(
                    'Length of {0} must be larger than {1}'.format(key, lmin))
        elif key == 'MaxOrder':
            value = int(value)
            lmax = min(len(getattr(self, k)) for k in polys)
            if not value < lmax:
                raise ValueError(
                    'MaxOrder must be smaller than {0}'.format(lmax))

        super(ThinMultipole, self).__setattr__(key, value)


class Multipole(LongElement, ThinMultipole):
    """pyAT multipole element"""
    REQUIRED_ATTRIBUTES = LongElement.REQUIRED_ATTRIBUTES + ['PolynomA',
                                                             'PolynomB']
    _conversions = dict(ThinMultipole._conversions, K=float, H=float)

    def __init__(self, family_name, length, poly_a, poly_b, **kwargs):
        """Multipole(FamName, Length, PolynomA, PolynomB, **keywords)

        Available keywords:
        MaxOrder        Number of desired multipoles. Default: highest index of
                        non-zero polynomial coefficients
        NumIntSteps     Number of integration steps (default: 10)
        KickAngle       Correction deviation angles (H, V)
        """
        kwargs.setdefault('PassMethod', 'StrMPoleSymplectic4Pass')
        kwargs.setdefault('NumIntSteps', 10)
        super(Multipole, self).__init__(family_name, length,
                                        poly_a, poly_b, **kwargs)


class Dipole(Multipole):
    """pyAT dipole element"""
    REQUIRED_ATTRIBUTES = LongElement.REQUIRED_ATTRIBUTES + ['BendingAngle',
                                                             'K']
    _conversions = dict(Multipole._conversions, EntranceAngle=float,
                        ExitAngle=float,
                        FringeInt1=float, FringeInt2=float,
                        FringeQuadEntrance=int, FringeQuadExit=int,
                        FringeBendEntrance=int, FringeBendExit=int)

    _entrance_fields = Multipole._entrance_fields + ['EntranceAngle',
                                                     'FringeInt1',
                                                     'FringeBendEntrance'
                                                     'FringeQuadEntrance']
    _exit_fields = Multipole._exit_fields + ['ExitAngle',
                                             'FringeInt2',
                                             'FringeBendExit',
                                             'FringeQuadExit']

    DefaultOrder = 0

    def __init__(self, family_name, length, bending_angle=0.0, k=0.0,
                 **kwargs):
        """Dipole(FamName, Length, bending_angle, Strength=0, **keywords)

        Available keywords:
        EntranceAngle   entrance angle (default 0.0)
        ExitAngle       exit angle (default 0.0)
        PolynomB        straight multipoles
        PolynomA        skew multipoles
        MaxOrder        Number of desired multipoles
        NumIntSteps     Number of integration steps (default: 10)
        FullGap         Magnet full gap
        FringeInt1      Fringe field extension
        FringeInt2
        FringeBendEntrance  1: legacy version Brown First Order (default)
                            2: SOLEIL close to second order of Brown
                            3: THOMX
        FringeBendExit
        FringeQuadEntrance  0: no fringe fiels effect (default)
                            1: Lee-Whiting's thin lens limit formula
                            2: elegant-like
        FringeQuadExit
        fringeIntM0     Integrals for FringeQuad method 2
        fringeIntP0
        KickAngle       Correction deviation angles (H, V)
        """
        poly_b = kwargs.pop('PolynomB', numpy.array([0, k]))
        kwargs.setdefault('BendingAngle', bending_angle)
        kwargs.setdefault('EntranceAngle', 0.0)
        kwargs.setdefault('ExitAngle', 0.0)
        kwargs.setdefault('PassMethod', 'BendLinearPass')
        super(Dipole, self).__init__(family_name, length, [], poly_b, **kwargs)

    def _part(self, fr, sumfr):
        pp = super(Dipole, self)._part(fr, sumfr)
        pp.BendingAngle = fr / sumfr * self.BendingAngle
        pp.EntranceAngle = 0.0
        pp.ExitAngle = 0.0
        return pp

    # noinspection PyPep8Naming
    @property
    def K(self):
        return self.PolynomB[1]

    # noinspection PyPep8Naming
    @K.setter
    def K(self, strength):
        self.PolynomB[1] = strength


# Bend is a synonym of Dipole.
Bend = Dipole


class Quadrupole(Multipole):
    """pyAT quadrupole element"""
    REQUIRED_ATTRIBUTES = LongElement.REQUIRED_ATTRIBUTES + ['K']
    _conversions = dict(Multipole._conversions, FringeQuadEntrance=int,
                        FringeQuadExit=int)

    _entrance_fields = Multipole._entrance_fields + ['FringeQuadEntrance']
    _exit_fields = Multipole._exit_fields + ['FringeQuadExit']

    DefaultOrder = 1

    def __init__(self, family_name, length, k=0.0, **kwargs):
        """Quadrupole(FamName, Length, Strength=0, **keywords)

        Available keywords:
        PolynomB        straight multipoles
        PolynomA        skew multipoles
        MaxOrder        Number of desired multipoles
        NumIntSteps     Number of integration steps (default: 10)
        FringeQuadEntrance  0: no fringe fiels effect (default)
                            1: Lee-Whiting's thin lens limit formula
                            2: elegant-like
        FringeQuadExit
        fringeIntM0     Integrals for FringeQuad method 2
        fringeIntP0
        KickAngle       Correction deviation angles (H, V)
        """
        poly_b = kwargs.pop('PolynomB', numpy.array([0, k]))
        kwargs.setdefault('PassMethod', 'QuadLinearPass')
        super(Quadrupole, self).__init__(family_name, length, [], poly_b,
                                         **kwargs)

    # noinspection PyPep8Naming
    @property
    def K(self):
        return self.PolynomB[1]

    # noinspection PyPep8Naming
    @K.setter
    def K(self, strength):
        self.PolynomB[1] = strength


class Sextupole(Multipole):
    """pyAT sextupole element"""
    REQUIRED_ATTRIBUTES = LongElement.REQUIRED_ATTRIBUTES + ['H']

    DefaultOrder = 2

    def __init__(self, family_name, length, h=0.0, **kwargs):
        """Sextupole(FamName, Length, Strength=0, **keywords)

        Available keywords:
        PolynomB        straight multipoles
        PolynomA        skew multipoles
        MaxOrder        Number of desired multipoles
        NumIntSteps     Number of integration steps (default: 10)
        KickAngle       Correction deviation angles (H, V)
        """
        poly_b = kwargs.pop('PolynomB', [0, 0, h])
        kwargs.setdefault('PassMethod', 'StrMPoleSymplectic4Pass')
        super(Sextupole, self).__init__(family_name, length, [], poly_b,
                                        **kwargs)

    # noinspection PyPep8Naming
    @property
    def H(self):
        return self.PolynomB[2]

    # noinspection PyPep8Naming
    @H.setter
    def H(self, strength):
        self.PolynomB[2] = strength


class Octupole(Multipole):
    """pyAT octupole element, with no changes from multipole at present"""
    REQUIRED_ATTRIBUTES = Multipole.REQUIRED_ATTRIBUTES

    DefaultOrder = 3


class RFCavity(LongElement):
    """pyAT RF cavity element"""
    REQUIRED_ATTRIBUTES = LongElement.REQUIRED_ATTRIBUTES + ['Voltage',
                                                             'Frequency',
                                                             'HarmNumber',
                                                             'Energy']
    _conversions = dict(LongElement._conversions,
                        Voltage=float, Frequency=float,
                        HarmNumber=int, TimeLag=float)

    def __init__(self, family_name, length, voltage, frequency,
                 harmonic_number, energy, **kwargs):
        """
        Available keywords:
        TimeLag   time lag with respect to the reference particle
        """
        kwargs.setdefault('TimeLag', 0.0)
        kwargs.setdefault('PassMethod', 'CavityPass')
        super(RFCavity, self).__init__(family_name, length,
                                       Voltage=voltage,
                                       Frequency=frequency,
                                       HarmNumber=harmonic_number,
                                       Energy=energy, **kwargs)

    def _part(self, fr, sumfr):
        pp = super(RFCavity, self)._part(fr, sumfr)
        pp.Voltage = fr * self.Voltage
        return pp


class M66(Element):
    """Linear (6, 6) transfer matrix"""
    REQUIRED_ATTRIBUTES = Element.REQUIRED_ATTRIBUTES
    _conversions = dict(Element._conversions, M66=_array66)

    def __init__(self, family_name, m66=None, **kwargs):
        if m66 is None:
            m66 = numpy.identity(6)
        kwargs.setdefault('PassMethod', 'Matrix66Pass')
        super(M66, self).__init__(family_name, M66=m66, **kwargs)


class Corrector(LongElement):
    """pyAT corrector element"""
    REQUIRED_ATTRIBUTES = LongElement.REQUIRED_ATTRIBUTES + ['KickAngle']

    def __init__(self, family_name, length, kick_angle, **kwargs):
        kwargs.setdefault('PassMethod', 'CorrectorPass')
        super(Corrector, self).__init__(family_name, length,
                                        KickAngle=kick_angle, **kwargs)


class Wiggler(LongElement):
    """pyAT wiggler element

    See atwiggler.m
    """
    REQUIRED_ATTRIBUTES = LongElement.REQUIRED_ATTRIBUTES + ['Lw', 'Bmax',
                                                             'Energy']
    _conversions = dict(Element._conversions, Lw=float, Bmax=float,
                        Energy=float,
                        Bx=lambda v: _array(v, (6, -1)),
                        By=lambda v: _array(v, (6, -1)),
                        Nstep=int, Nmeth=int, NHharm=int, NVharm=int)

    def __init__(self, family_name, length, wiggle_period, b_max, energy,
                 Nstep=5, Nmeth=4, By=(1, 1, 0, 1, 1, 0), Bx=(), **kwargs):
        """
        Args:
            length: total length of the wiggler
            wiggle_period: length must be a multiple of this
            b_max: peak wiggler field [Tesla]
            energy: beam energy [eV]

        Available keywords:
            Nstep: number of integration steps.
            Nmeth: symplectic integration order: 2 or 4
            Bx: harmonics for horizontal wiggler: (6,nHharm) array-like object
            By: harmonics for vertical wiggler (6,nHharm) array-like object

        """
        kwargs.setdefault('PassMethod', 'GWigSymplecticPass')
        n_wiggles = length / wiggle_period
        if abs(round(n_wiggles) - n_wiggles) > 1e-6:
            raise ValueError("Wiggler: length / wiggle_period is not an "
                             "integer. ({0}/{1}={2})".format(length,
                                                             wiggle_period,
                                                             n_wiggles))
        super(Wiggler, self).__init__(family_name, length, Lw=wiggle_period,
                                      Bmax=b_max, Nstep=Nstep, Nmeth=Nmeth,
                                      By=By, Bx=Bx, Energy=energy, **kwargs)

        for i, b in enumerate(self.By.T):
            dk = abs(b[3]**2 - b[4]**2 - b[2]**2) / abs(b[4])
            if dk > 1e-6:
                raise ValueError("Wiggler(H): kx^2 + kz^2 -ky^2 !=0, i = "
                                 "{0}".format(i))

        for i, b in enumerate(self.Bx.T):
            dk = abs(b[2]**2 - b[4]**2 - b[3]**2) / abs(b[4])
            if dk > 1e-6:
                raise ValueError("Wiggler(V): ky^2 + kz^2 -kx^2 !=0, i = "
                                 "{0}".format(i))

        self.NHharm = self.By.shape[1]
        self.NVharm = self.Bx.shape[1]


CLASS_MAP = dict((k, v) for k, v in locals().items()
                 if isinstance(v, type) and issubclass(v, Element))



"""Python port of the Accelerator Toolbox"""

# Make all functions visible in the at namespace:
from .lattice import *
from .tracking import *
from .physics import *
from .load import *
from .matching import *



"""
Collection of functions to fit various global parameters
such as the tune and the chromaticity
"""
import numpy
from at.lattice import set_value_refpts
from at.physics import linopt

__all__ = ['fit_tune', 'fit_chrom']


def _get_tune(ring, dp=0):
    _, tune, _, _ = linopt(ring, dp=dp)
    return tune


def _get_chrom(ring, dp=0):
    _, _, chrom, _ = linopt(ring, dp=dp, get_chrom=True)
    return chrom


def _fit_tune_chrom(ring, order, func, refpts1, refpts2, newval, tol=1.0e-12,
                    dp=0, niter=3):

    def _get_resp(ring, order, func, refpts, attname, delta, dp=0):
        set_value_refpts(ring, refpts, attname, delta, order=order,
                         increment=True)
        datap = func(ring, dp=dp)
        set_value_refpts(ring, refpts, attname, -2*delta, order=order,
                         increment=True)
        datan = func(ring, dp=dp)
        set_value_refpts(ring, refpts, attname, delta, order=order,
                         increment=True)
        data = numpy.subtract(datap, datan)/(2*delta)
        return data

    delta = 1e-6*10**(order)
    val = func(ring, dp=dp)
    dq1 = _get_resp(ring, order, func, refpts1, 'PolynomB', delta, dp=dp)
    dq2 = _get_resp(ring, order, func, refpts2, 'PolynomB', delta, dp=dp)
    J = [[dq1[0], dq2[0]], [dq1[1], dq2[1]]]
    dk = numpy.linalg.solve(J, numpy.subtract(newval, val))
    set_value_refpts(ring, refpts1, 'PolynomB', dk[0], order=order,
                     increment=True)
    set_value_refpts(ring, refpts2, 'PolynomB', dk[1], order=order,
                     increment=True)

    val = func(ring, dp=dp)
    sumsq = numpy.sum(numpy.square(numpy.subtract(val, newval)))
    if sumsq > tol and niter > 0:
        _fit_tune_chrom(ring, order, func, refpts1, refpts2, newval, tol=tol,
                        dp=dp, niter=niter-1)
    else:
        return


def fit_tune(ring, refpts1, refpts2, newval, tol=1.0e-12, dp=0, niter=3):
    """
    Function to fit the tune of the ring, using 2 families defined by
    refpts

    Args:
        ring: lattice for which the tune needs to be matched
        refpts1/2: refpts for the 2 families
        newval: new tunes
        tol: tolerance for the matching [default=1.0e-12]
        dp: dp/p at which the values need to be matched [default=0]
        niter: maximum number of iterations to reach tol [default=3]

    Typical usage:
    at.matching.fit_tune(ring, refpts1, refpts2, [0.1,0.25])
    """
    _fit_tune_chrom(ring, 1, _get_tune, refpts1, refpts2, newval, tol=tol,
                    dp=dp, niter=niter)


def fit_chrom(ring, refpts1, refpts2, newval, tol=1.0e-12, dp=0, niter=3):
    """
    Function to fit the chromaticity of the ring, using 2 families
    defined by refpts

    Args:
        ring: lattice for which the chromaticity needs to be matched
        ring: lattice for which the chromaticity needs to be matched
        refpts1/2: refpts for the 2 families
        newval: new tunes
        tol: tolerance for the matching [default=1.0e-12]
        dp: dp/p at which the values need to be matched [default=0]
        niter: maximum number of iterations to reach tol [default=3]

    Typical usage:
    at.matching.fit_chrom(ring, refpts1, refpts2, [10,5])
    """
    _fit_tune_chrom(ring, 2, _get_chrom, refpts1, refpts2, newval, tol=tol,
                    dp=dp, niter=niter)



from math import pi, sqrt, asin, cos
import numpy
from numpy import nan
from scipy.constants import c
from at.lattice import Lattice
from at.physics import Cgamma, Cq, e_mass

__all__ = ['RingParameters', 'radiation_parameters', 'envelope_parameters']


class RingParameters(object):
    """Class for pretty printing the ring properties"""

    props = {
        'tunes':            '              Frac. tunes: {0}',
        'tunes6':           '  Frac. tunes (6D motion): {0}',
        'fulltunes':        '                    Tunes: {0}',
        'chromaticities':   '           Chromaticities: {0}',
        'alphac':           ' Momentum compact. factor: {0:e}',
        'etac':             '              Slip factor: {0:e}',
        'E0':               '                   Energy: {0:e} eV',
        'U0':               '       Energy loss / turn: {0:e} eV',
        'i1':               ' Radiation integrals - I1: {0} m',
        'i2':               '                       I2: {0} m^-1',
        'i3':               '                       I3: {0} m^-2',
        'i4':               '                       I4: {0} m^-1',
        'i5':               '                       I5: {0} m^-1',
        'emittances':       '          Mode emittances: {0}',
        'J':                'Damping partition numbers: {0}',
        'Tau':              '            Damping times: {0} s',
        'sigma_e':          '            Energy spread: {0:g}',
        'sigma_l':          '             Bunch length: {0:g} m',
        'voltage':          '         Cavities voltage: {0} V',
        'phi_s':            '        Synchrotron phase: {0:g} rd',
        'f_s':              '    Synchrotron frequency; {0:g} Hz'
    }

    def __init__(self, **kwargs):
        """Initialisation"""
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __str__(self):
        vrs = vars(self).copy()
        vals = [(self.props[k], vrs.pop(k, None)) for k in self.props]
        # Predefined attributes
        lines = [f.format(v) for f, v in vals if v is not None]
        # Other attributes
        lines += ['{0:>25}: {1}'.format(k, getattr(self, k)) for k in vrs]
        return '\n'.join(lines)


# noinspection PyPep8Naming
def radiation_parameters(ring, dp=0.0, params=None):
    """Compute ring parameters from the radiation integrals. Valid for
    uncoupled lattices with no RF cavity or radiating element.

    INPUT
        ring            Lattice object.
        dp=0.0          momentum deviation.

    KEYWORD
        params=None     RingParam object to be updated.

    OUTPUT
        params          RingParam object. The computed attributes are,

            tunes           (3,) fractional (H, V, Long.) tunes
            fulltunes       (3,) full tunes
            chromaticities  (2,) H, V Chromaticities
            alphac          Momentum compaction factor
            etac            Frequency slip factor
            E0              Energy [eV]
            U0              nergy loss / turn [eV]
            i1              Radiation integrals - I1 [m]
            i2                                    I2 [m^-1]
            i3                                    I3 [m^-2]
            i4                                    I4 [m^-1]
            i5                                    I5 [m^-1]
            emittances      (3,) Mode emittances
            J               (3,) Damping partition numbers
            Tau             (3,) Damping times [s]
            sigma_e         Energy spread
            sigma_l         Bunch length [m]
            voltage         Total accelerating voltage [V]
            phi_s           Synchrotron phase [rad]
            f_s             Synchrotron frequency [Hz]
    """
    rp = RingParameters() if params is None else params
    _, tunes, chroms, twiss = ring.linopt(dp, range(len(ring) + 1),
                                          get_chrom=True, coupled=False)
    rp.chromaticities = chroms * ring.periodicity
    integs = ring.get_radiation_integrals(dp, twiss=twiss)
    rp.i1, rp.i2, rp.i3, rp.i4, rp.i5 = numpy.array(integs) * ring.periodicity
    circumference = ring.circumference
    revolution_period = circumference / c
    voltage = ring.voltage
    E0 = ring.energy
    gamma = E0 / e_mass
    gamma2 = gamma * gamma
    beta = sqrt(1.0 - 1.0/gamma2)
    U0 = Cgamma / 2.0 / pi * E0**4 * rp.i2
    Jx = 1.0 - rp.i4/rp.i2
    Jz = 1.0
    Je = 2.0 + rp.i4/rp.i2
    damping_partition_numbers = numpy.array([Jx, Jz, Je])
    ct = 2.0 * E0 / U0 * revolution_period
    rp.E0 = E0
    rp.U0 = U0
    emitx = Cq * gamma2 * rp.i5 / Jx / rp.i2
    rp.emittances = numpy.array([emitx, nan, nan])
    alphac = rp.i1 / circumference
    etac = 1.0/gamma2 - alphac
    rp.phi_s = (pi - asin(U0 / voltage)) if U0 <= voltage else nan
    nus = sqrt(abs(etac * ring.harmonic_number *
               voltage * cos(rp.phi_s) / E0 / 2.0 / pi)) / beta
    rp.voltage = voltage
    rp.f_s = nus / revolution_period
    rp.Tau = ct / damping_partition_numbers
    rp.J = damping_partition_numbers
    rp.sigma_e = sqrt(Cq * gamma2 * rp.i3 / Je / rp.i2)
    rp.sigma_l = beta * abs(etac) * circumference / 2.0 / pi / nus * rp.sigma_e
    ringtunes, _ = numpy.modf(ring.periodicity * tunes)
    rp.fulltunes = ring.periodicity * twiss[-1].mu / 2.0 / pi
    rp.tunes = numpy.concatenate((ringtunes, (nus,)))
    rp.alphac = alphac
    rp.etac = etac
    return rp


# noinspection PyPep8Naming
def envelope_parameters(ring, params=None):
    """Compute ring parameters from ohmi_envelope

    INPUT
        ring            Lattice object.

    KEYWORD
        params=None     RingParam object to be updated.

    OUTPUT
        params          RingParam object. The computed attributes are,

            tunes6          (3,) fractional (H, V, Long.) tunes (6D motion)
            emittances      (3,) Mode emittances
            J               (3,) Damping partition numbers
            Tau             (3,) Damping times [s]
            sigma_e         Energy spread
            sigma_l         Bunch length [m]
            voltage         Total accelerating voltage [V]
            phi_s           Synchrotron phase [rad]
            f_s             Synchrotron frequency [Hz]
    """
    rp = RingParameters() if params is None else params
    emit0, beamdata, emit = ring.ohmi_envelope()
    voltage = ring.voltage
    rp.E0 = ring.energy
    rp.U0 = ring.energy_loss
    revolution_period = ring.circumference / c
    rp.Tau = revolution_period / beamdata.damping_rates / ring.periodicity
    alpha = 1.0 / rp.Tau
    rp.J = 4.0 * alpha / numpy.sum(alpha)
    rp.tunes6, _ = numpy.modf(ring.periodicity * beamdata.tunes)
    rp.phi_s = pi - asin(rp.U0 / voltage)
    rp.voltage = voltage
    rp.f_s = rp.tunes6[2] / revolution_period
    rp.emittances = beamdata.mode_emittances
    rp.sigma_e = sqrt(emit0.r66[4, 4])
    rp.sigma_l = sqrt(emit0.r66[5, 5])
    return rp


Lattice.radiation_parameters = radiation_parameters
Lattice.envelope_parameters = envelope_parameters



"""Global set of constants for numerical differentiation"""


class _Dst(object):
    """Set of constants for AT numerical analysis"""
    # optimized values:
    # see https://github.com/atcollab/at/pull/199#issuecomment-745420443
    XYStep = 3.e-8           # Coordinate step for differentiation
    DPStep = 3.e-6           # Momentum step for dispersion and chromaticity
    OrbConvergence = 1.e-12  # Convergence criterion for orbit
    OrbMaxIter = 20          # Max. number of iterations for orbit

    def __setattr__(self, name, value):
        _ = getattr(self, name)     # make sure attribute exists
        object.__setattr__(self, name, value)

    def reset(self, name):
        delattr(self, name)


DConstant = _Dst()



"""Generic function to save and load python AT lattices. The format is
determined by the file extension
"""
import os.path
from at.lattice import Lattice

__all__ = ['load_lattice', 'save_lattice', 'register_format']

_load_extension = {}
_save_extension = {}


def load_lattice(filepath, **kwargs):
    """Load a Lattice object from a file

    The file format is indicated by the filepath extension.

    PARAMETERS
        filepath        name of the file

    KEYWORDS
        name            Name of the lattice
                        (default: taken from the file, or '')
        energy          Energy of the lattice
                        (default: taken from the file)
        periodicity     Number of periods
                        (default: taken from the file, or 1)
        *               all other keywords will be set as Lattice attributes

    MAT-FILE SPECIFIC KEYWORDS
        mat_key         name of the Matlab variable containing the lattice.
                        Default: Matlab variable name if there is only one,
                        otherwise 'RING'
        check=True      if False, skip the coherence tests
        quiet=False     If True, suppress the warning for non-standard classes
        keep_all=False  if True, keep RingParam elements as Markers

    Known extensions are:
    """
    _, ext = os.path.splitext(filepath)
    try:
        return _load_extension[ext.lower()](filepath, **kwargs)
    except KeyError:
        print("Could not load lattice file with extension {}.".format(ext))


def save_lattice(ring, filepath, **kwargs):
    """Save a Lattice object

    The file format is indicated by the filepath extension.

    PARAMETERS
        ring            Lattice object
        filepath        name of the file

    MAT-FILE SPECIFIC KEYWORDS
        mat_key='RING'  Name of the Matlab variable

    Known extensions are:
    """
    _, ext = os.path.splitext(filepath)
    try:
        return _save_extension[ext.lower()](ring, filepath, **kwargs)
    except KeyError:
        print("Could not save lattice file with extension {}.".format(ext))


def register_format(extension, load_func=None, save_func=None, descr=''):
    """Register format-specific processing functions

    PARAMETERS
        extension       File extension string
        load_func       load function (default: None)
        save_func       save_lattice function (default: None)
        descr           File type description
    """
    if load_func is not None:
        _load_extension[extension] = load_func
        load_lattice.__doc__ += '\n    {0:<10}\t{1}'.format(extension, descr)
    if save_func is not None:
        _save_extension[extension] = save_func
        save_lattice.__doc__ += '\n    {0:<10}\t{1}'.format(extension, descr)


Lattice.load = staticmethod(load_lattice)
Lattice.save = save_lattice



import sys
import numpy
import pytest
from at import elements
from at.lattice import Lattice, AtWarning, AtError


def test_lattice_creation_gets_attributes_from_arguments():
    lat = Lattice(name='lattice', energy=3.e+6, periodicity=32, an_attr=12)
    assert len(lat) == 0
    assert lat.name == 'lattice'
    assert lat.energy == 3.e+6
    assert lat.periodicity == 32
    assert lat._radiation is False
    assert lat.an_attr == 12


def test_lattice_creation_short_scan_reads_radiation_status_correctly():
    d = elements.Dipole('d1', 1, BendingAngle=numpy.pi, Energy=5.e+6,
                        PassMethod='BndMPoleSymplectic4RadPass')
    lat = Lattice([d], name='lattice', energy=3.e+6, periodicity=32)
    assert len(lat) == 1
    assert lat.name == 'lattice'
    assert lat.energy == 3.e+6
    assert lat.periodicity == 32
    assert lat._radiation is True


def test_lattice_creation_from_lattice_inherits_attributes():
    d = elements.Dipole('d1', 1, BendingAngle=numpy.pi, Energy=5.e+6,
                        PassMethod='BndMPoleSymplectic4RadPass')
    lat1 = Lattice([d], name='lattice', energy=3.e+6, periodicity=32,
                   an_attr=12)
    lat1.another_attr = 5
    lat2 = Lattice(lat1)
    assert id(lat1) != id(lat2)
    assert len(lat2) == 1
    assert lat2.name == 'lattice'
    assert lat2.energy == 3.e+6
    assert lat2.periodicity == 32
    assert lat2._radiation is True
    assert lat2.an_attr == 12
    assert lat2.another_attr == 5


def test_lattice_energy_is_not_defined_raises_AtError():
    with pytest.raises(AtError):
        Lattice()


def test_item_is_not_an_AT_element_warns_correctly():
    with pytest.warns(AtWarning):
        Lattice(['a'], energy=0, periodicity=1)


def test_lattice_string_ordering():
    lat = Lattice([elements.Drift('D0', 1.0, attr1=numpy.array(0))],
                  name='lat', energy=5, periodicity=1, attr2=3)
    # Default dictionary ordering is only in Python >= 3.6
    if sys.version_info < (3, 6):
        assert lat.__str__().startswith("Lattice(<1 elements>, ")
        assert lat.__str__().endswith(", attr2=3)")
        assert lat.__repr__().startswith("Lattice([Drift('D0', 1.0, "
                                         "attr1=array(0))], ")
        assert lat.__repr__().endswith(", attr2=3)")
    else:
        assert lat.__str__() == ("Lattice(<1 elements>, name='lat', energy=5,"
                                 " periodicity=1, attr2=3)")
        assert lat.__repr__() == ("Lattice([Drift('D0', 1.0, attr1=array(0))],"
                                  " name='lat', energy=5, periodicity=1,"
                                  " attr2=3)")


def test_getitem(simple_lattice, simple_ring):
    assert simple_lattice[-1] == simple_ring[-1]
    bool_refs = numpy.array([False, True, False, True, False, True])
    assert simple_lattice[bool_refs] == simple_ring[1::2]


def test_setitem(simple_lattice):
    new = elements.Monitor('M2')
    old = simple_lattice[5]
    simple_lattice[5] = new
    assert simple_lattice[5] != old
    assert simple_lattice[5] == new
    bool_refs = numpy.array([False, False, False, False, False, True])
    simple_lattice[bool_refs] = old
    assert simple_lattice[5] != new
    assert simple_lattice[5] == old


def test_delitem(simple_lattice, simple_ring):
    mon = elements.Monitor('M2')
    simple_lattice.append(mon)
    assert len(simple_lattice) == 7
    del simple_lattice[-1]
    assert simple_lattice[:] == simple_ring
    bool_refs = numpy.array([False, False, False, False, False, False, True])
    simple_lattice.append(mon)
    assert len(simple_lattice) == 7
    del simple_lattice[bool_refs]
    assert simple_lattice[:] == simple_ring


def test_copy(hmba_lattice):
    assert id(hmba_lattice.copy()) != id(hmba_lattice)
    assert id(hmba_lattice.copy()[0]) == id(hmba_lattice[0])


def test_deepcopy(hmba_lattice):
    assert id(hmba_lattice.deepcopy()) != id(hmba_lattice)
    assert id(hmba_lattice.deepcopy()[0]) != id(hmba_lattice[0])


def test_property_values_against_known(hmba_lattice):
    assert hmba_lattice.voltage == 6000000
    assert hmba_lattice.harmonic_number == 992
    assert hmba_lattice.radiation is False
    numpy.testing.assert_almost_equal(hmba_lattice.energy_loss,
                                      2526188.658758993, decimal=1)


def test_radiation_change(hmba_lattice):
    rfs = [elem for elem in hmba_lattice if isinstance(elem,
                                                       elements.RFCavity)]
    dipoles = [elem for elem in hmba_lattice if isinstance(elem,
                                                           elements.Dipole)]
    quads = [elem for elem in hmba_lattice if isinstance(elem,
                                                         elements.Quadrupole)]
    hmba_lattice.radiation_on(None, 'pass2', 'auto')
    assert hmba_lattice.radiation is True
    for elem in rfs:
        assert elem.PassMethod == 'IdentityPass'
    for elem in dipoles:
        assert elem.PassMethod == 'pass2'
    for elem in quads:
        assert elem.PassMethod == 'StrMPoleSymplectic4RadPass'
    hmba_lattice.radiation_off(None, 'BndMPoleSymplectic4Pass', 'auto')
    assert hmba_lattice.radiation is False
    for elem in rfs:
        assert elem.PassMethod == 'IdentityPass'
    for elem in dipoles:
        assert elem.PassMethod == 'BndMPoleSymplectic4Pass'
    for elem in quads:
        assert elem.PassMethod == 'StrMPoleSymplectic4Pass'


def test_radiation_state_errors(hmba_lattice):
    hmba_lattice.radiation_on()
    with pytest.raises(AtError):
        hmba_lattice.linopt()
    hmba_lattice.radiation_off()
    hmba_lattice.linopt()
    with pytest.raises(AtError):
        hmba_lattice.ohmi_envelope()
    hmba_lattice.radiation_on()
    hmba_lattice.ohmi_envelope()
    with pytest.raises(AtError):
        hmba_lattice.get_mcf()
    hmba_lattice.radiation_off()
    hmba_lattice.get_mcf()



"""
Conversion utilities for creating pyat elements
"""
import collections
import os
import re
import numpy
from warnings import warn
from distutils import sysconfig
from at import integrators
from at.lattice import AtWarning
from at.lattice import CLASS_MAP, elements as elt
# imports necessary in' globals()' for 'eval'
# noinspection PyUnresolvedReferences
from numpy import array, uint8  # For global namespace


class RingParam(elt.Element):
    """Private class for Matlab RingParam element"""
    REQUIRED_ATTRIBUTES = elt.Element.REQUIRED_ATTRIBUTES + ['Energy',
                                                             'Periodicity']
    _conversions = dict(elt.Element._conversions, Energy=float, Periodicity=int)

    def __init__(self, family_name, energy, periodicity=1, **kwargs):
        kwargs.setdefault('Periodicity', periodicity)
        kwargs.setdefault('PassMethod', 'IdentityPass')
        super(RingParam, self).__init__(family_name, Energy=energy, **kwargs)


_alias_map = {'rbend': elt.Dipole,
              'sbend': elt.Dipole,
              'quad': elt.Quadrupole,
              'sext': elt.Sextupole,
              'rf': elt.RFCavity,
              'bpm': elt.Monitor,
              'ap': elt.Aperture,
              'ringparam': RingParam,
              'wig': elt.Wiggler}


# Matlab to Python class translation
_CLASS_MAP = dict((k.lower(), v) for k, v in CLASS_MAP.items())
_CLASS_MAP.update(_alias_map)

_PASS_MAP = {'DriftPass': elt.Drift,
             'BendLinearPass': elt.Dipole,
             'BndMPoleSymplectic4RadPass': elt.Dipole,
             'BndMPoleSymplectic4Pass': elt.Dipole,
             'QuadLinearPass': elt.Quadrupole,
             'CorrectorPass': elt.Corrector,
             'CavityPass': elt.RFCavity, 'RFCavityPass': elt.RFCavity,
             'ThinMPolePass': elt.ThinMultipole,
             'Matrix66Pass': elt.M66,
             'AperturePass': elt.Aperture,
             'GWigSymplecticPass': elt.Wiggler}

# Matlab to Python attribute translation
_param_to_lattice = {'Energy': 'energy', 'Periodicity': 'periodicity',
                     'FamName': 'name'}

# Python to Matlab class translation
_matclass_map = {'Dipole': 'Bend'}

# Python to Matlab type translation
_mattype_map = {int: float,
                numpy.ndarray: lambda attr: numpy.asanyarray(attr,
                                                             dtype=float)}

_class_to_matfunc = {
    elt.Dipole: 'atsbend',
    elt.Bend: 'atsbend',
    elt.M66: 'atM66'}


def hasattrs(kwargs, *attributes):
    """Check if the element would have the specified attribute(s), i.e. if they
    are in kwargs; allows checking for multiple attributes in one go.

    Args:
        kwargs (dict): The dictionary of keyword arguments passed to the
                        Element constructor.
        attributes (iterable): A list of strings, the attribute names to be
                                checked.

    Returns:
        bool: A single boolean, True if the element has any of the specified
               attributes.
    """
    for attribute in attributes:
        if attribute in kwargs:
            return True
    return False


def find_class(elem_dict, quiet=False):
    """Attempts to correctly identify the Class of the element from its kwargs.

    Args:
        elem_dict       The dictionary of keyword arguments passed to the
                        Element constructor.
    Keywords:
        quiet=False     If True, suppress the warning for non-standard classes

    Returns:
        element_class:  The guessed Class name
    """

    def low_order(key):
        polynom = numpy.array(elem_dict[key], dtype=numpy.float64).reshape(-1)
        try:
            low = numpy.where(polynom != 0.0)[0][0]
        except IndexError:
            low = -1
        return low

    class_name = elem_dict.pop('Class', '')
    try:
        return _CLASS_MAP[class_name.lower()]
    except KeyError:
        if not quiet and class_name:
            warn(AtWarning("Class '{0}' does not exist.\n"
                           "{1}".format(class_name, elem_dict)))
        fam_name = elem_dict.get('FamName', '')
        try:
            return _CLASS_MAP[fam_name.lower()]
        except KeyError:
            pass_method = elem_dict.get('PassMethod', '')
            if not quiet and not pass_method:
                warn(AtWarning("No PassMethod provided."
                               "\n{0}".format(elem_dict)))
            elif not quiet and not pass_method.endswith('Pass'):
                warn(AtWarning("Invalid PassMethod ({0}), provided pass "
                               "methods should end in 'Pass'."
                               "\n{1}".format(pass_method, elem_dict)))
            class_from_pass = _PASS_MAP.get(pass_method)
            if class_from_pass is not None:
                return class_from_pass
            else:
                length = float(elem_dict.get('Length', 0.0))
                if hasattrs(elem_dict, 'FullGap', 'FringeInt1', 'FringeInt2',
                            'gK', 'EntranceAngle', 'ExitAngle'):
                    return elt.Dipole
                elif hasattrs(elem_dict, 'Voltage', 'Frequency', 'HarmNumber',
                              'PhaseLag', 'TimeLag'):
                    return elt.RFCavity
                elif hasattrs(elem_dict, 'Periodicity'):
                    # noinspection PyProtectedMember
                    return RingParam
                elif hasattrs(elem_dict, 'Limits'):
                    return elt.Aperture
                elif hasattrs(elem_dict, 'M66'):
                    return elt.M66
                elif hasattrs(elem_dict, 'K'):
                    return elt.Quadrupole
                elif hasattrs(elem_dict, 'PolynomB', 'PolynomA'):
                    loworder = low_order('PolynomB')
                    if loworder == 1:
                        return elt.Quadrupole
                    elif loworder == 2:
                        return elt.Sextupole
                    elif loworder == 3:
                        return elt.Octupole
                    elif (pass_method.startswith('StrMPoleSymplectic4') or
                          (length > 0)):
                        return elt.Multipole
                    else:
                        return elt.ThinMultipole
                elif hasattrs(elem_dict, 'KickAngle'):
                    return elt.Corrector
                elif length > 0.0:
                    return elt.Drift
                elif hasattrs(elem_dict, 'GCR'):
                    return elt.Monitor
                else:
                    return elt.Marker


def get_pass_method_file_name(pass_method):
    extension_list = sysconfig.get_config_vars('EXT_SUFFIX', 'SO')
    extension = set(filter(None, extension_list))
    ext = extension.pop() if len(extension) == 1 else '.so'
    return pass_method + ext


def element_from_dict(elem_dict, index=None, check=True, quiet=False):
    """return an AT element from a dictionary of attributes
    """

    # noinspection PyShadowingNames
    def sanitise_class(index, cls, elem_dict):
        """Checks that the Class and PassMethod of the element are a valid
            combination. Some Classes and PassMethods are incompatible and
            would raise errors during calculation if left, so we raise an error
            here with a more helpful message.

        Args:
            index:          element index
            cls:            Proposed class
            elem_dict:      he dictionary of keyword arguments passed to the
                            Element constructor.

        Raises:
            AttributeError: if the PassMethod and Class are incompatible.
        """
        def err(message, *args):
            location = ': ' if index is None else ' {0}: '.format(index)
            msg = ''.join(('Error in element', location,
                           'PassMethod {0} '.format(pass_method),
                           message.format(*args), '\n{0}'.format(elem_dict)))
            return AttributeError(msg)

        class_name = cls.__name__
        pass_method = elem_dict.get('PassMethod')
        if pass_method is not None:
            pass_to_class = _PASS_MAP.get(pass_method)
            length = float(elem_dict.get('Length', 0.0))
            file_name = get_pass_method_file_name(pass_method)
            file_path = os.path.join(integrators.__path__[0], file_name)
            if not os.path.isfile(os.path.realpath(file_path)):
                raise err("does not have a {0} file.".format(file_name))
            elif (pass_method == 'IdentityPass') and (length != 0.0):
                raise err("is not compatible with length {0}.", length)
            elif pass_to_class is not None:
                if not issubclass(cls, pass_to_class):
                    raise err("is not compatible with Class {0}.", class_name)
            elif issubclass(cls, (elt.Marker, elt.Monitor, RingParam)):
                if pass_method != 'IdentityPass':
                    raise err("is not compatible with Class {0}.", class_name)
            elif cls == elt.Drift:
                if pass_method != 'DriftPass':
                    raise err("is not compatible with Class {0}.", class_name)

    cls = find_class(elem_dict, quiet=quiet)
    if check:
        sanitise_class(index, cls, elem_dict)
    # Remove mandatory attributes from the keyword arguments.
    # Create list rather than generator to ensure that elements are removed
    # from elem_dict.
    elem_args = [elem_dict.pop(attr, None) for attr in cls.REQUIRED_ATTRIBUTES]
    element = cls(*(arg for arg in elem_args if arg is not None), **elem_dict)
    return element


def element_from_string(elem_string):
    """Generate an AT-element from its repr string"""
    return eval(elem_string, globals(), CLASS_MAP)


def element_from_m(line):
    """Generate a AT-element from a line in a m-file"""

    def convert(argmt):
        """convert Matlab syntax to numpy syntax"""

        def setarray(arr):
            lns = arr.split(';')
            rr = [setarray(v) for v in lns] if len(lns) > 1 else lns[0].split()
            return '[{0}]'.format(', '.join(rr))

        if argmt.startswith('['):
            return 'array({0})'.format(setarray(argmt[1:-1]))
        else:
            return argmt

    left = line.index('(')
    right = line.rindex(')')
    cls = _CLASS_MAP[line[:left].strip()[2:]]
    arguments = [a.strip() for a in line[left + 1:right].split(',')]
    ll = len(cls.REQUIRED_ATTRIBUTES)
    if ll < len(arguments) and arguments[ll].endswith("Pass'"):
        arguments.insert(ll, "'PassMethod'")
    keys = arguments[ll::2]
    vals = arguments[ll + 1::2]
    args = [convert(v) for v in arguments[:ll]]
    keywords = ['='.join((k[1:-1], convert(v))) for k, v in zip(keys, vals)]
    elem_string = '{0}({1})'.format(cls.__name__, ', '.join(args + keywords))
    return element_from_string(elem_string)


def element_to_dict(elem):
    """Generate the Matlab structure for a AT element"""
    dct = dict((k, _mattype_map.get(type(v), lambda attr: attr)(v))
               for k, v in elem.items())
    class_name = elem.__class__.__name__
    dct['Class'] = _matclass_map.get(class_name, class_name)
    return dct


def element_to_m(elem):
    """Generate the Matlab-evaluable string for a AT element"""

    def convert(arg):
        if isinstance(arg, numpy.ndarray):
            if arg.ndim > 1:
                lns = (str(list(ln)).replace(',', '')[1:-1] for ln in arg)
                return ''.join(('[', '; '.join(lns), ']'))
            elif arg.ndim > 0:
                return str(list(arg)).replace(',', '')
            else:
                return str(arg)
        else:
            return repr(arg)

    def m_name(elclass):
        stdname = ''.join(('at', elclass.__name__.lower()))
        return _class_to_matfunc.get(elclass, stdname)

    attrs = dict(elem.items())
    args = [attrs.pop(k, getattr(elem, k)) for k in elem.REQUIRED_ATTRIBUTES]
    defelem = elem.__class__(*args)
    kwds = dict((k, v) for k, v in attrs.items()
                if not numpy.array_equal(v, getattr(defelem, k, None)))
    argstrs = [convert(arg) for arg in args]
    if 'PassMethod' in kwds:
        argstrs.append(convert(kwds.pop('PassMethod')))
    argstrs += [', '.join((repr(k), convert(v))) for k, v in kwds.items()]
    return '{0:>15}({1});...'.format(m_name(elem.__class__), ', '.join(argstrs))


# Kept for compatibility but should be deprecated:


CLASS_MAPPING = dict((key, cls.__name__) for (key, cls) in _CLASS_MAP.items())

PASS_MAPPING = dict((key, cls.__name__) for (key, cls) in _PASS_MAP.items())


def find_class_name(elem_dict, quiet=False):
    return find_class(elem_dict, quiet=quiet).__name__


def split_ignoring_parentheses(string, delimiter):
    PLACEHOLDER = "placeholder"
    substituted = string[:]
    matches = collections.deque(re.finditer("\\(.*?\\)", string))
    for match in matches:
        substituted = substituted.replace(match.group(), PLACEHOLDER, 1)
    parts = substituted.split(delimiter)
    replaced_parts = []
    for part in parts:
        if PLACEHOLDER in part:
            next_match = matches.popleft()
            part = part.replace(PLACEHOLDER, next_match.group(), 1)
        replaced_parts.append(part)
    assert not matches

    return replaced_parts



"""
Helper functions for working with AT lattices.

A lattice as understood by pyAT is any sequence of elements.  These functions
are useful for working with these sequences.

The refpts allow functions to select points in the lattice, returned values are
given at the entrance of each element specified in refpts;
refpts can be:
    - an integer in the range [-len(ring), len(ring)-1] selecting the element
        according to python indexing rules. As a special case, len(ring) is
        allowed and refers to the end of the last element,
    - an ordered list of such integers without duplicates,
    - a numpy array of booleans of maximum length len(ring)+1, where selected
        elements are True.
"""
import numpy
import functools
from fnmatch import fnmatch
from at.lattice import elements

__all__ = ['AtError', 'AtWarning', 'check_radiation', 'uint32_refpts',
           'bool_refpts', 'checkattr', 'checktype', 'get_cells',
           'get_elements', 'get_refpts', 'refpts_iterator', 'get_s_pos',
           'set_shift', 'set_tilt', 'tilt_elem', 'shift_elem',
           'get_value_refpts', 'set_value_refpts']


class AtError(Exception):
    pass


class AtWarning(UserWarning):
    pass


def check_radiation(rad):
    """Function to be used as a decorator for optics functions

    If ring is a Lattice object, raises an exception
        if ring.radiation is not rad

    If ring is any other sequence, no test is performed
    """
    def radiation_decorator(func):
        @functools.wraps(func)
        def wrapper(ring, *args, **kwargs):
            ringrad = getattr(ring, 'radiation', rad)
            if ringrad != rad:
                raise AtError('{0} needs radiation {1}'.format(
                    func.__name__, 'ON' if rad else 'OFF'))
            return func(ring, *args, **kwargs)
        return wrapper
    return radiation_decorator


def uint32_refpts(refpts, n_elements):
    """
    Return a uint32 numpy array with contents as the indices of the selected
    elements.  This is used for indexing a lattice using explicit indices.
    """
    refs = numpy.ravel(refpts)
    if (refpts is None) or (refs.size == 0):
        return numpy.array([], dtype=numpy.uint32)
    elif numpy.issubdtype(refs.dtype, numpy.bool_):
        return numpy.flatnonzero(refs).astype(numpy.uint32)

    # Handle negative indices
    refs = numpy.array([i if (i == n_elements) else i % n_elements
                        for i in refs], dtype=numpy.uint32)

    # Check ascending
    if refs.size > 1:
        prev = refs[0]
        for nxt in refs[1:]:
            if nxt < prev:
                raise ValueError('refpts should be given in ascending order')
            elif nxt == prev:
                raise ValueError('refpts contains duplicates or index(es) out'
                                 ' of range')
            prev = nxt

    return refs


def bool_refpts(refpts, n_elements):
    """
    Return a boolean numpy array of length n_elements + 1 where True elements
    are selected. This is used for indexing a lattice using True or False
    values.
    """
    refs = numpy.ravel(refpts)
    if (refpts is None) or (refs.size == 0):
        return numpy.zeros(n_elements + 1, dtype=bool)
    elif refs.dtype == bool:
        diff = 1 + n_elements - refs.size
        if diff <= 0:
            return refs[:n_elements + 1]
        else:
            return numpy.append(refs, numpy.zeros(diff, dtype=bool))
    else:
        brefpts = numpy.zeros(n_elements + 1, dtype=bool)
        brefpts[refs] = True
        return brefpts


def checkattr(*args):
    """Return a function to be used as a filter. Check the presence or the
    value of an attribute. This function can be used to extract from a ring
    all elements have a given attribute.

    filtfunc = checkattr(attrname)
        returns the function filtfunc such that ok=filtfunc(element) is True if
        the element has a 'attrname' attribute

    filtfunc = checkattr(attrname, attrvalue)
        returns the function filtfunc such that ok=filtfunc(element) is True if
        the element has a 'attrname' attribute with the value attrvalue

    Examples:

    cavs = filter(checkattr('Frequency'), ring)
        returns an iterator over all elements in ring that have a
        'Frequency' attribute

    elts = filter(checkattr('K', 0.0), ring)
        returns an iterator over all elements in ring that have a 'K'
        attribute equal to 0.0
    """

    def testf(el):
        try:
            v = getattr(el, args[0])
            return len(args) == 1 or v == args[1]
        except AttributeError:
            return False

    return testf


def checktype(eltype):
    """Return a function to be used as a filter. Check the type of an element.
    This function can be used to extract from a ring all elements have a
    given type.

    filtfunc = checktype(class)
        returns the function filtfunc such that ok=filtfunc(element) is True
        if the element is an instance of class

    Example:

    qps = filter(checktype(at.Quadrupole), ring)
        returns an iterator over all quadrupoles in ring
    """
    return lambda el: isinstance(el, eltype)


def checkname(pattern):
    """Return a function to be used as a filter. Check the name of an element.
    This function can be used to extract from a ring all elements have a
    given FamName attribute.

    filtfunc = checkname(pattern)
        returns the function filtfunc such that ok=filtfunc(element) is True
        if the element's FamName matches pattern.

    Example:

    qps = filter(checkname('QF.*'), ring)
        returns an iterator over all elements
    """
    return lambda el: fnmatch(el.FamName, pattern)


def get_cells(ring, *args):
    """Return a numpy array of booleans, with the same length as ring,
    marking all elements satisfying a given condition.

    refpts = getcells(ring, filtfunc)
        selects all elements for which the function filtfunc(element)
        returns True

    refpts = getcells(ring, attrname)
        selects all elements having a 'attrname' attribute

    refpts = getcells(ring, attrname, attrvalue)
        selects all elements having a 'attrname' attribute with value attrvalue

    Example:

    refpts = getcells(ring, 'Frequency')
        returns a numpy array of booleans where all elements having a
        'Frequency' attribute are True

    refpts = getcells(ring, 'K', 0.0)
        returns a numpy array of booleans where all elements having a 'K'
        attribute equal to 0.0 are True
    """
    if callable(args[0]):
        testf = args[0]
    else:
        testf = checkattr(*args)
    return numpy.array(tuple(map(testf, ring)), dtype=bool)


def refpts_iterator(ring, refpts):
    """Return an iterator over all elements in ring identified by refpts.

    refpts may be:

    1) a list of integers (0 indicating the first element)
    2) a numpy array of booleans as long as ring where selected elements
       are true
    """
    refs = numpy.ravel(refpts)
    if (refpts is None) or (refs.size == 0):
        return
    elif refs.dtype == bool:
        for el, tst in zip(ring, refs):
            if tst:
                yield el
    else:
        for i in refs:
            yield ring[i]


def refpts_count(refpts, n_elements):
    refs = numpy.ravel(refpts)
    if (refpts is None) or (refs.size == 0):
        return 0
    elif refs.dtype == bool:
        return numpy.count_nonzero(refs)
    else:
        return len(refs)


def get_refpts(ring, key, quiet=True):
    """Get the elements refpts of a family or class (type) from the lattice.

    Args:
        ring: lattice from which to retrieve the elements.
        key: can be:
             1) an element instance, will return all elements of the same type
                in the lattice, e.g. key=Drift('d1', 1.0)
             2) an element type, will return all elements of that type in the
                lattice, e.g. key=at.elements.Sextupole
             3) a string to match against elements' FamName, supports Unix
                shell-style wildcards, e.g. key='BPM_*1'
        quiet: if false print information about matched elements for FamName
               matches, defaults to True.

    Returns:
        elems: a list of elems refpts matching key
    """
    if isinstance(key, elements.Element):
        checkfun = checktype(type(key))
    elif isinstance(key, type):
        checkfun = checktype(key)
    elif numpy.issubdtype(type(key), numpy.str_):
        checkfun = checkname(key)
        if not quiet:
            matched_fams = set(elem.FamName for elem in filter(checkfun, ring))
            ending = 'y' if len(matched_fams) == 1 else 'ies'
            print("String '{0}' matched {1} famil{2}: {3}\n"
                  "all corresponding elements have been "
                  "returned.".format(key, len(matched_fams), ending,
                                     ', '.join(matched_fams)))
    else:
        raise TypeError("Invalid key type {0}; please enter a string, element"
                        " type, or element instance.".format(type(key)))
    return uint32_refpts(list(map(checkfun, ring)), len(ring))


def get_elements(ring, key, quiet=True):
    """Get the elements of a family or class (type) from the lattice.

    Args:
        ring: lattice from which to retrieve the elements.
        key: can be:
             1) an element instance, will return all elements of the same type
                in the lattice, e.g. key=Drift('d1', 1.0)
             2) an element type, will return all elements of that type in the
                lattice, e.g. key=at.elements.Sextupole
             3) a string to match against elements' FamName, supports Unix
                shell-style wildcards, e.g. key='BPM_*1'
        quiet: if false print information about matched elements for FamName
               matches, defaults to True.

    Returns:
        a list of elems matching key
    """
    if isinstance(key, elements.Element):
        checkfun = checktype(type(key))
    elif isinstance(key, type):
        checkfun = checktype(key)
    elif numpy.issubdtype(type(key), numpy.str_):
        checkfun = checkname(key)
        if not quiet:
            matched_fams = set(elem.FamName for elem in filter(checkfun, ring))
            ending = 'y' if len(matched_fams) == 1 else 'ies'
            print("String '{0}' matched {1} famil{2}: {3}\n"
                  "all corresponding elements have been "
                  "returned.".format(key, len(matched_fams), ending,
                                     ', '.join(matched_fams)))
    else:
        raise TypeError("Invalid key type {0}; please enter a string, element"
                        " type, or element instance.".format(type(key)))
    return list(filter(checkfun, ring))


def get_value_refpts(ring, refpts, var, index=None, order=None):
    """Get the values of an attribute of an array of elements based on
    their refpts

    Args:
        ring: lattice from which to retrieve the elements.
        refpts: integer or array of integer or booleans
        var: attribute name
        index (optional): index of the value to change in case var is
        an array, if None the full array is returned (Default)
    """
    if order is not None:   # for backward compatibility with the
        index = order       # obsolete keyword 'order'

    if index is None:
        def getf(elem):
            return getattr(elem, var)
    else:
        def getf(elem):
            return getattr(elem, var)[index]

    return numpy.array([getf(elem) for elem in refpts_iterator(ring, refpts)])


def set_value_refpts(ring, refpts, var, values, index=None, order=None,
                     increment=False):
    """Set the values of an attribute of an array of elements based on
    their refpts

    Args:
        ring: lattice from which to retrieve the elements.
        refpts: integer or array of integer or booleans
        var: attribute name
        values: desired value for the attribute
        index (optional): index of the value to change in case var is an
        array, if None the full array is replaced by value (Default)
        increment (optional): adds value to the initial value, if False
        the initial value is replaced (Default)
    """
    if order is not None:   # for backward compatibility with the
        index = order       # obsolete keyword 'order'

    if index is None:
        def setf(elem, value):
            setattr(elem, var, value)
    else:
        def setf(elem, value):
            getattr(elem, var)[index] = value

    if increment:
        values = values + get_value_refpts(ring, refpts, var, index=index)
    else:
        values = values + numpy.zeros(refpts_count(refpts, len(ring)))

    for elm, val in zip(refpts_iterator(ring, refpts), values):
        setf(elm, val)


def get_s_pos(ring, refpts=None):
    """
    Return a numpy array corresponding to the s position of the specified
    elements.

    Args:
        ring: lattice from which to retrieve s position
        refpts: elements at which to return s position. If None, return
                s position at all elements in the ring.
    """
    if refpts is None:
        refpts = range(len(ring) + 1)
    # Positions at the end of each element.
    s_pos = numpy.cumsum([getattr(el, 'Length', 0.0) for el in ring])
    # Prepend position at the start of the first element.
    s_pos = numpy.concatenate(([0.0], s_pos))
    refpts = uint32_refpts(refpts, len(ring))
    return s_pos[refpts]


def tilt_elem(elem, rots, relative=False):
    """
    set a new tilt angle to an element.
    The rotation matrices are stored in the R1 and R2 attributes

    R1 = [[ cos(rots) sin(rots)]    R2 = [[cos(rots) -sin(rots)]
          [-sin(rots) cos(rots)]]         [sin(rots)  cos(rots)]]

    elem            element to be tilted
    rots            tilt angle (in radians).
                    rots > 0 corresponds to a corkskew rotation of the element
                    looking in the direction of the beam
    relative=False  Rotation relative to the previous element rotation
    """
    cs = numpy.cos(rots)
    sn = numpy.sin(rots)
    rm = numpy.asfortranarray(numpy.diag([cs, cs, cs, cs, 1.0, 1.0]))
    rm[0, 2] = sn
    rm[1, 3] = sn
    rm[2, 0] = -sn
    rm[3, 1] = -sn
    if relative and hasattr(elem, 'R1') and hasattr(elem, 'R2'):
        elem.R1 = elem.R1.dot(rm)
        elem.R2 = rm.T.dot(elem.R2)
    else:
        elem.R1 = rm
        elem.R2 = rm.T


def shift_elem(elem, deltax=0.0, deltaz=0.0, relative=False):
    """
    set a new displacement vector to an element.
    The ranslation vectors are stored in the T1 and T2 attributes

    elem            element to be displaced
    deltax          horizontal displacement of the element
    deltaz          vertical displacement of the element
    relative=False  Displacement relative to the previous alignment
    """
    tr = numpy.array([deltax, 0.0, deltaz, 0.0, 0.0, 0.0])
    if relative and hasattr(elem, 'T1') and hasattr(elem, 'T2'):
        elem.T1 -= tr
        elem.T2 += tr
    else:
        elem.T1 = -tr
        elem.T2 = tr


def set_tilt(ring, tilts, relative=False):
    """Set tilts to a list of elements.

    ring            sequence of elements to be tilted
    tilts           sequence of tilt values as long as ring or
                    scalar tilt value applied to all elements
    relative=False  Rotation relative to the previous tilt angle
    """
    tilts = numpy.broadcast_to(tilts, (len(ring),))
    for el, tilt in zip(ring, tilts):
        tilt_elem(el, tilt, relative=relative)


def set_shift(ring, dxs, dzs, relative=False):
    """Set translations to a list of elements.

    ring            sequence of elements to be shifted
    dxs             sequence of horizontal displacement as long as ring or
                    scalar horizontal displacement applied to all elements
    dzs             sequence of vertical displacement as long as ring or
                    scalar vertical displacement applied to all elements
    relative=False  Displacement relative to the previous alignment
    """
    dxs = numpy.broadcast_to(dxs, (len(ring),))
    dzs = numpy.broadcast_to(dzs, (len(ring),))
    for el, dx, dy in zip(ring, dxs, dzs):
        shift_elem(el, dx, dy, relative=relative)



"""
Coupled or non-coupled 4x4 linear motion
"""
import numpy
from math import sqrt, atan2, pi
from at.lattice import Lattice, check_radiation, uint32_refpts, get_s_pos, \
    bool_refpts
from at.tracking import lattice_pass
from at.physics import find_orbit4, find_m44, jmat, DConstant
from .harmonic_analysis import get_tunes_harmonic

__all__ = ['linopt', 'avlinopt', 'get_mcf', 'get_tune',
           'get_chrom']

DDP = 1e-8

_jmt = jmat(1)

# dtype for structured array containing Twiss parameters
TWISS_DTYPE = [('idx', numpy.uint32),
               ('s_pos', numpy.float64),
               ('closed_orbit', numpy.float64, (6,)),
               ('dispersion', numpy.float64, (4,)),
               ('alpha', numpy.float64, (2,)),
               ('beta', numpy.float64, (2,)),
               ('mu', numpy.float64, (2,)),
               ('m44', numpy.float64, (4, 4))]

# dtype for structured array containing linopt parameters
LINDATA_DTYPE = TWISS_DTYPE + [('A', numpy.float64, (2, 2)),
                               ('B', numpy.float64, (2, 2)),
                               ('C', numpy.float64, (2, 2)),
                               ('gamma', numpy.float64)]


def _twiss22(ms, alpha0, beta0):
    """
    Calculate Twiss parameters from the standard 2x2 transfer matrix
    (i.e. x or y).
    """
    bbb = ms[:, 0, 1]
    aaa = ms[:, 0, 0] * beta0 - bbb * alpha0
    beta = (aaa * aaa + bbb * bbb) / beta0
    alpha = -(aaa * (ms[:, 1, 0] * beta0 - ms[:, 1, 1] * alpha0) +
              bbb * ms[:, 1, 1]) / beta0
    mu = numpy.arctan2(bbb, aaa)
    # Unwrap negative jumps in betatron phase advance
    dmu = numpy.diff(numpy.append([0], mu))
    jumps = dmu < 0
    mu += numpy.cumsum(jumps) * 2.0 * numpy.pi
    return alpha, beta, mu


def _closure(m22):
    diff = (m22[0, 0] - m22[1, 1]) / 2.0
    sinmu = numpy.sign(m22[0, 1]) * sqrt(-m22[0, 1] * m22[1, 0] - diff * diff)
    cosmu = 0.5 * numpy.trace(m22)
    alpha = diff / sinmu
    beta = m22[0, 1] / sinmu
    tune = (atan2(sinmu, cosmu) / 2.0 / pi) % 1
    return alpha, beta, tune


def _chromfunc(ddp, *param_up_down):
    l0_up, tune_up, _, l_up, l0_down, tune_down, _, l_down = param_up_down
    db0 = (l0_up['beta'] - l0_down['beta']) / ddp
    db = (l_up['beta'] - l_down['beta']) / ddp
    da0 = (l0_up['alpha'] - l0_down['alpha']) / ddp
    da = (l_up['alpha'] - l_down['alpha']) / ddp
    chrom = (tune_up - tune_down) / ddp
    dispersion = (l_up['closed_orbit'] -
                  l_down['closed_orbit'])[:, :4] / ddp
    return chrom, dispersion, db0, da0, db, da


# noinspection PyPep8Naming
@check_radiation(False)
def linopt(ring, dp=0.0, refpts=None, get_chrom=False, orbit=None,
           keep_lattice=False, coupled=True, twiss_in=None, **kwargs):
    """
    Perform linear analysis of a lattice
    lindata0, tune, chrom, lindata = linopt(ring, dp[, refpts])
    PARAMETERS
        ring            lattice description.
        dp=0.0          momentum deviation.
        refpts=None     elements at which data is returned. It can be:
                        1) an integer in the range [-len(ring), len(ring)-1]
                           selecting the element according to python indexing
                           rules. As a special case, len(ring) is allowed and
                           refers to the end of the last element,
                        2) an ordered list of such integers without duplicates,
                        3) a numpy array of booleans of maximum length
                           len(ring)+1, where selected elements are True.
    KEYWORDS
        orbit           avoids looking for the closed orbit if is already known
                        ((6,) array)
        get_chrom=False compute dispersion and chromaticities. Needs computing
                        the optics at 2 different momentum deviations around
                        the central one.
        keep_lattice    Assume no lattice change since the previous tracking.
                        Defaults to False
        XYStep=1.0e-8   transverse step for numerical computation
        DPStep=1.0E-6   momentum deviation used for computation of
                        chromaticities and dispersion
        coupled=True    if False, simplify the calculations by assuming
                        no H/V coupling
        twiss_in=None   Initial twiss to compute transfer line optics of the
                        type lindata, the initial orbit in twiss_in is ignored,
                        only the beta and alpha are required other quatities
                        set to 0 if absent
    OUTPUT
        lindata0        linear optics data at the entrance/end of the ring
        tune            [tune_A, tune_B], linear tunes for the two normal modes
                        of linear motion [1]
        chrom           [ksi_A , ksi_B], chromaticities ksi = d(nu)/(dP/P).
                        Only computed if 'get_chrom' is True
        lindata         linear optics at the points refered to by refpts, if
                        refpts is None an empty lindata structure is returned.

        lindata is a record array with fields:
        idx             element index in the ring
        s_pos           longitudinal position [m]
        closed_orbit    (6,) closed orbit vector
        dispersion      (4,) dispersion vector
                        Only computed if 'get_chrom' is True
        m44             (4, 4) transfer matrix M from the beginning of ring
                        to the entrance of the element [2]
        mu              [mux, muy], betatron phase (modulo 2*pi)
        beta            [betax, betay] vector
        alpha           [alphax, alphay] vector
        All values given at the entrance of each element specified in refpts.
        In case coupled = True additional outputs are available:
        A               (2, 2) matrix A in [3]
        B               (2, 2) matrix B in [3]
        C               (2, 2) matrix C in [3]
        gamma           gamma parameter of the transformation to eigenmodes
        Field values can be obtained with either
        lindata['idx']    or
        lindata.idx
    REFERENCES
        [1] D.Edwars,L.Teng IEEE Trans.Nucl.Sci. NS-20, No.3, p.885-888, 1973
        [2] E.Courant, H.Snyder
        [3] D.Sagan, D.Rubin Phys.Rev.Spec.Top.-Accelerators and beams,
            vol.2 (1999)
    """

    # noinspection PyShadowingNames
    def analyze(r44):
        t44 = r44.reshape((4, 4))
        mm = t44[:2, :2]
        nn = t44[2:, 2:]
        m = t44[:2, 2:]
        n = t44[2:, :2]
        gamma = sqrt(numpy.linalg.det(numpy.dot(n, C) + numpy.dot(G, nn)))
        msa = (G.dot(mm) - m.dot(_jmt.dot(C.T.dot(_jmt.T)))) / gamma
        msb = (numpy.dot(n, C) + numpy.dot(G, nn)) / gamma
        cc = (numpy.dot(mm, C) + numpy.dot(G, m)).dot(
            _jmt.dot(msb.T.dot(_jmt.T)))
        return msa, msb, gamma, cc

    xy_step = kwargs.pop('XYStep', DConstant.XYStep)
    dp_step = kwargs.pop('DPStep', DConstant.DPStep)
    uintrefs = uint32_refpts([] if refpts is None else refpts, len(ring))

    # Get initial orbit
    if twiss_in is None:
        if orbit is None:
            orbit, _ = find_orbit4(ring, dp, keep_lattice=keep_lattice,
                                   XYStep=xy_step)
            keep_lattice = True
        disp0 = numpy.NaN
        if get_chrom:
            orbit_up, _ = find_orbit4(ring, dp + 0.5*dp_step, XYStep=xy_step,
                                      keep_lattice=keep_lattice)
            orbit_down, _ = find_orbit4(ring, dp - 0.5*dp_step, XYStep=xy_step,
                                        keep_lattice=keep_lattice)
            disp0 = numpy.array(orbit_up - orbit_down)[:4] / dp_step
    else:
        if orbit is None:
            orbit = numpy.zeros((6,))
        disp0 = numpy.NaN
        if get_chrom:
            try:
                disp0 = twiss_in['dispersion']
            except KeyError:
                print('Dispersion not found in twiss_in, setting to zero')
                disp0 = numpy.zeros((4,))
            dorbit = numpy.hstack((0.5 * dp_step * disp0,
                                   numpy.array([0.5 * dp_step, 0])))
            orbit_up = orbit+dorbit
            orbit_down = orbit-dorbit

    orbs = numpy.squeeze(
        lattice_pass(ring, orbit.copy(order='K'), refpts=uintrefs,
                     keep_lattice=keep_lattice), axis=(1, 3)).T
    m44, mstack = find_m44(ring, dp, uintrefs, orbit=orbit, keep_lattice=True,
                           XYStep=xy_step)

    nrefs = uintrefs.size

    M = m44[:2, :2]
    N = m44[2:, 2:]
    m = m44[:2, 2:]
    n = m44[2:, :2]

    # Calculate A, B, C, gamma at the first element
    if coupled:
        H = m + _jmt.dot(n.T.dot(_jmt.T))
        t = numpy.trace(M - N)
        t2 = t * t
        t2h = t2 + 4.0 * numpy.linalg.det(H)

        g = sqrt(1.0 + sqrt(t2 / t2h)) / sqrt(2.0)
        G = numpy.diag((g, g))
        C = -H * numpy.sign(t) / (g * sqrt(t2h))
        A = G.dot(G.dot(M)) - numpy.dot(G, (
            m.dot(_jmt.dot(C.T.dot(_jmt.T))) + C.dot(n))) + C.dot(
            N.dot(_jmt.dot(C.T.dot(_jmt.T))))
        B = G.dot(G.dot(N)) + numpy.dot(G, (
            _jmt.dot(C.T.dot(_jmt.T.dot(m))) + n.dot(C))) + _jmt.dot(
            C.T.dot(_jmt.T.dot(M.dot(C))))
    else:
        A = M
        B = N
        C = numpy.zeros((2, 2))
        g = 1.0

    # Get initial twiss parameters
    if twiss_in is None:
        a0_a, b0_a, tune_a = _closure(A)
        a0_b, b0_b, tune_b = _closure(B)
        tune = numpy.array([tune_a, tune_b])
    else:
        try:
            a0_a, a0_b = twiss_in['alpha'][0], twiss_in['alpha'][1]
        except KeyError:
            raise ValueError('Initial alpha required for transfer line')
        try:
            b0_a, b0_b = twiss_in['beta'][0], twiss_in['beta'][1]
        except KeyError:
            raise ValueError('Initial beta required for transfer line')
        try:
            tune = numpy.array([twiss_in['mu'][0], twiss_in['mu'][1]])/(2*pi)
        except KeyError:
            print('Mu not found in twiss_in, setting to zero')
            tune = numpy.zeros((2,))

    if get_chrom:
        # noinspection PyUnboundLocalVariable
        kwup = dict(orbit=orbit_up, twiss_in=twiss_in)
        # noinspection PyUnboundLocalVariable
        kwdown = dict(orbit=orbit_down, twiss_in=twiss_in)
        param_up = linopt(ring, dp=dp + 0.5*dp_step, refpts=uintrefs,
                          keep_lattice=True, coupled=coupled,
                          XYStep=xy_step, **kwup)
        param_down = linopt(ring, dp - 0.5*dp_step, refpts=uintrefs,
                            keep_lattice=True, coupled=coupled,
                            XYStep=xy_step, **kwdown)
        param_up_down = param_up+param_down
        chrom, dispersion, _, _, _, _ = _chromfunc(dp_step, *param_up_down)
    else:
        chrom = numpy.array([numpy.NaN, numpy.NaN])
        dispersion = numpy.NaN

    lindata0 = numpy.rec.fromarrays(
        (len(ring), get_s_pos(ring, len(ring))[0], orbit, disp0,
         numpy.array([a0_a, a0_b]), numpy.array([b0_a, b0_b]),
         2.0 * pi * tune, m44, A, B, C, g),
        dtype=LINDATA_DTYPE)

    # Propagate to reference points
    lindata = numpy.rec.array(numpy.zeros(nrefs, dtype=LINDATA_DTYPE))
    if nrefs > 0:
        if coupled:
            MSA, MSB, gamma, CL = zip(*[analyze(ms44) for ms44 in mstack])
            msa = numpy.stack(MSA, axis=0)
            msb = numpy.stack(MSB, axis=0)
            AL = [ms.dot(A.dot(_jmt.dot(ms.T.dot(_jmt.T))))
                  for ms in msa]
            BL = [ms.dot(B.dot(_jmt.dot(ms.T.dot(_jmt.T))))
                  for ms in msb]
        else:
            msa = mstack[:, :2, :2]
            msb = mstack[:, 2:, 2:]
            AL = numpy.NaN
            BL = numpy.NaN
            CL = numpy.NaN
            gamma = numpy.NaN

        alpha_a, beta_a, mu_a = _twiss22(msa, a0_a, b0_a)
        alpha_b, beta_b, mu_b = _twiss22(msb, a0_b, b0_b)

        if twiss_in is not None:
            qtmp = numpy.array([mu_a[-1], mu_b[-1]])/(2 * numpy.pi)
            qtmp -= numpy.floor(qtmp)
            mu_a += tune[0]*2*pi
            mu_b += tune[1]*2*pi
            tune = qtmp

        lindata['idx'] = uintrefs
        lindata['s_pos'] = get_s_pos(ring, uintrefs)
        lindata['closed_orbit'] = orbs
        lindata['m44'] = mstack
        lindata['alpha'] = numpy.stack((alpha_a, alpha_b), axis=1)
        lindata['beta'] = numpy.stack((beta_a, beta_b), axis=1)
        lindata['dispersion'] = dispersion
        lindata['mu'] = numpy.stack((mu_a, mu_b), axis=1)
        lindata['A'] = AL
        lindata['B'] = BL
        lindata['C'] = CL
        lindata['gamma'] = gamma

    return lindata0, tune, chrom, lindata


# noinspection PyPep8Naming
@check_radiation(False)
def avlinopt(ring, dp=0.0, refpts=None, **kwargs):
    """
    Perform linear analysis of a lattice and returns average beta, dispersion
    and phase advance

    lindata,avebeta,avemu,avedisp,tune,chrom = avlinopt(ring, dp[, refpts])

    PARAMETERS
        ring            lattice description.
        dp=0.0          momentum deviation.
        refpts=None     elements at which data is returned. It can be:
                        1) an integer in the range [-len(ring), len(ring)-1]
                           selecting the element according to python indexing
                           rules. As a special case, len(ring) is allowed and
                           refers to the end of the last element,
                        2) an ordered list of such integers without duplicates,
                        3) a numpy array of booleans of maximum length
                           len(ring)+1, where selected elements are True.

    KEYWORDS
        orbit           avoids looking for the closed orbit if is already known
                        ((6,) array)
        keep_lattice    Assume no lattice change since the previous tracking.
                        Defaults to False
        XYStep=1.0e-8   transverse step for numerical computation
        DPStep=1.0E-8   momentum deviation used for computation of
                        chromaticities and dispersion
        coupled=True    if False, simplify the calculations by assuming
                        no H/V coupling

    OUTPUT
        lindata         linear optics at the points refered to by refpts, if
                        refpts is None an empty lindata structure is returned.
                        See linopt for details
        avebeta         Average beta functions [betax,betay] at refpts
        avemu           Average phase advances [mux,muy] at refpts
        avedisp         Average dispersion [Dx,Dx',Dy,Dy',muy] at refpts
        avespos         Average s position at refpts
        tune            [tune_A, tune_B], linear tunes for the two normal modes
                        of linear motion [1]
        chrom           [ksi_A , ksi_B], chromaticities ksi = d(nu)/(dP/P).
                        Only computed if 'get_chrom' is True


    See also get_twiss,linopt

    """
    def get_strength(elem):
        try:
            k = elem.PolynomB[1]
        except (AttributeError, IndexError):
            k = 0.0
        return k

    def betadrift(beta0, beta1, alpha0, lg):
        gamma0 = (alpha0 * alpha0 + 1) / beta0
        return 0.5 * (beta0 + beta1) - gamma0 * lg * lg / 6

    def betafoc(beta1, alpha0, alpha1, k2, lg):
        gamma1 = (alpha1 * alpha1 + 1) / beta1
        return 0.5 * ((gamma1 + k2 * beta1) * lg + alpha1 - alpha0) / k2 / lg

    def dispfoc(dispp0, dispp1, k2, lg):
        return (dispp0 - dispp1) / k2 / lg

    boolrefs = bool_refpts([] if refpts is None else refpts, len(ring))
    length = numpy.array([el.Length for el in ring[boolrefs]])
    strength = numpy.array([get_strength(el) for el in ring[boolrefs]])
    longelem = bool_refpts([], len(ring))
    longelem[boolrefs] = (length != 0)

    shorti_refpts = (~longelem) & boolrefs
    longi_refpts = longelem & boolrefs
    longf_refpts = numpy.roll(longi_refpts, 1)

    all_refs = shorti_refpts | longi_refpts | longf_refpts
    _, tune, chrom, d_all = linopt(ring, dp=dp, refpts=all_refs,
                                   get_chrom=True, **kwargs)
    lindata = d_all[boolrefs[all_refs]]

    avebeta = lindata.beta.copy()
    avemu = lindata.mu.copy()
    avedisp = lindata.dispersion.copy()
    aves = lindata.s_pos.copy()

    di = d_all[longi_refpts[all_refs]]
    df = d_all[longf_refpts[all_refs]]

    long = (length != 0.0)
    kfoc = (strength != 0.0)
    foc = long & kfoc
    nofoc = long & (~kfoc)
    K2 = numpy.stack((strength[foc], -strength[foc]), axis=1)
    fff = foc[long]
    length = length.reshape((-1, 1))

    avemu[long] = 0.5 * (di.mu + df.mu)
    aves[long] = 0.5 * (df.s_pos + di.s_pos)
    avebeta[nofoc] = \
        betadrift(di.beta[~fff], df.beta[~fff], di.alpha[~fff], length[nofoc])
    avebeta[foc] = \
        betafoc(df.beta[fff], di.alpha[fff], df.alpha[fff], K2, length[foc])
    avedisp[numpy.ix_(long, [1, 3])] = \
        (df.dispersion[:, [0, 2]] - di.dispersion[:, [0, 2]]) / length[long]
    idx = numpy.ix_(~fff, [0, 2])
    avedisp[numpy.ix_(nofoc, [0, 2])] = (di.dispersion[idx] +
                                         df.dispersion[idx]) * 0.5
    idx = numpy.ix_(fff, [1, 3])
    avedisp[numpy.ix_(foc, [0, 2])] = \
        dispfoc(di.dispersion[idx], df.dispersion[idx], K2, length[foc])
    return lindata, avebeta, avemu, avedisp, aves, tune, chrom


@check_radiation(False)
def get_mcf(ring, dp=0.0, keep_lattice=False, **kwargs):
    """Compute momentum compaction factor

    PARAMETERS
        ring            lattice description
        dp              momentum deviation. Defaults to 0

    KEYWORDS
        keep_lattice    Assume no lattice change since the previous tracking.
                        Defaults to False
        dp_step=1.0E-6  momentum deviation used for differentiation
    """
    dp_step = kwargs.pop('DPStep', DConstant.DPStep)
    fp_a, _ = find_orbit4(ring, dp=dp - 0.5*dp_step, keep_lattice=keep_lattice)
    fp_b, _ = find_orbit4(ring, dp=dp + 0.5*dp_step, keep_lattice=True)
    fp = numpy.stack((fp_a, fp_b),
                     axis=0).T  # generate a Fortran contiguous array
    b = numpy.squeeze(lattice_pass(ring, fp, keep_lattice=True), axis=(2, 3))
    ring_length = get_s_pos(ring, len(ring))
    return (b[5, 1] - b[5, 0]) / dp_step / ring_length[0]


@check_radiation(False)
def get_tune(ring, method='linopt', dp=0.0, **kwargs):
    """
    gets the tune using several available methods

    method can be 'linopt', 'fft' or 'laskar'
    linopt: returns the tune from the linopt function
    fft: tracks a single particle (one for x and one for y)
    and computes the tune from the fft
    laskar: tracks a single particle (one for x and one for y)
    and computes the harmonic components


    INPUT
    for linopt:
        no input needed

    for harmonic:
        nturns: number of turn
        amplitude: amplitude of oscillation
        method: laskar or fft
        num_harmonics: number of harmonic components to compute
        (before mask applied, default=20)
        fmin/fmax: determine the boundaries within which the tune is
        located [default 0->1]
        hann: flag to turn on hanning window [default-> False]
        remove_dc: Removes the mean offset of oscillation data

    OUTPUT
        tunes = np.array([Qx,Qy])
    """

    def gen_centroid(ring, ampl, nturns, dp, remove_dc):
        orbit, _ = find_orbit4(ring, dp)
        ld, _, _, _ = linopt(ring, dp, orbit=orbit)

        invx = numpy.array([[1/numpy.sqrt(ld['beta'][0]), 0],
                            [ld['alpha'][0]/numpy.sqrt(ld['beta'][0]),
                            numpy.sqrt(ld['beta'][0])]])

        invy = numpy.array([[1/numpy.sqrt(ld['beta'][1]), 0],
                            [ld['alpha'][1]/numpy.sqrt(ld['beta'][1]),
                            numpy.sqrt(ld['beta'][1])]])

        p0 = numpy.array([orbit, ]*2).T
        p0[0, 0] += ampl
        p0[2, 1] += ampl
        p1 = lattice_pass(ring, p0, refpts=len(ring), nturns=nturns)
        cent_x = p1[0, 0, 0, :]
        cent_xp = p1[1, 0, 0, :]
        cent_y = p1[2, 1, 0, :]
        cent_yp = p1[3, 1, 0, :]
        if remove_dc:
            cent_x -= numpy.mean(cent_x)
            cent_y -= numpy.mean(cent_y)
            cent_xp -= numpy.mean(cent_xp)
            cent_yp -= numpy.mean(cent_yp)

        cent_x, cent_xp = numpy.matmul(invx, [cent_x, cent_xp])
        cent_y, cent_yp = numpy.matmul(invy, [cent_y, cent_yp])
        return (cent_x - 1j * cent_xp,
                cent_y - 1j * cent_yp)

    if method == 'linopt':
        _, tunes, _, _ = linopt(ring, dp=dp)
    else:
        nturns = kwargs.pop('nturns', 512)
        ampl = kwargs.pop('ampl', 1.0e-6)
        remove_dc = kwargs.pop('remove_dc', True)
        cent_x, cent_y = gen_centroid(ring, ampl, nturns, dp, remove_dc)
        cents = numpy.vstack((cent_x, cent_y))
        tunes = get_tunes_harmonic(cents, method, **kwargs)
    return tunes


@check_radiation(False)
def get_chrom(ring, method='linopt', dp=0, **kwargs):
    """
    gets the chromaticity using several available methods

    method can be 'linopt', 'fft' or 'laskar'
    linopt: returns the chromaticity from the linopt function
    fft: tracks a single particle (one for x and one for y)
    and computes the tune from the fft
    harmonic: tracks a single particle (one for x and one for y)
    and computes the harmonic components

    see get_tune for kwargs inputs

    OUTPUT
        chromaticities = np.array([Q'x,Q'y])
    """

    dp_step = kwargs.pop('DPStep', DConstant.DPStep)
    if method == 'fft':
        print('Warning fft method not accurate to get the ' +
              'chromaticity')

    tune_up = get_tune(ring, method=method, dp=dp + 0.5 * dp_step, **kwargs)
    tune_down = get_tune(ring, method=method, dp=dp - 0.5 * dp_step, **kwargs)
    chrom = (tune_up - tune_down) / dp_step
    return numpy.array(chrom)


Lattice.linopt = linopt
Lattice.avlinopt = avlinopt
Lattice.get_mcf = get_mcf
Lattice.avlinopt = avlinopt
Lattice.get_tune = get_tune
Lattice.get_chrom = get_chrom



"""
Helper functions for working with AT lattices.

A lattice as understood by pyAT is any sequence of elements.  These functions
are useful for working with these sequences.
"""

from .elements import *
from .utils import *
from .lattice_object import *
